#!/usr/bin/env python
import numpy as np


TAU = 2 * np.pi  # References: tauday.com


def draw_expanded_scales(imgL, sel_kpts, exkpts, exdesc_):
    import plottool as pt
    draw_keypoint_patch = pt.draw_keypoint_patch
    get_warped_patch = pt.get_warped_patch  # NOQA

    # Rows are for different scales
    # Cols are for different patches
    # There is a prefix row showing the original image
    nRows, nCols = len(exkpts), len(sel_kpts)
    exkpts_ = np.vstack(exkpts)

    fnum = 1
    pt.figure(fnum=fnum, docla=True, doclf=True)

    nPreRows = 1
    nPreCols = (nPreRows * nCols) + 1

    pt.show_keypoints(imgL, exkpts_, fnum=fnum, pnum=(nRows + nPreRows, 1, 1),
                      color=pt.BLUE)

    px = 0
    for row, kpts_ in enumerate(exkpts):
        for col, kp in enumerate(kpts_):
            sift = exdesc_[px]
            pnum = (nRows + nPreRows, nCols, px + nPreCols)
            draw_keypoint_patch(imgL, kp, sift, warped=True, fnum=fnum, pnum=pnum)
            px += 1
    #df2.draw()

    print('nRows = %r' % nRows)
    print('nCols = %r' % nCols)


def in_depth_ellipse(kp):
    """
    Makes sure that I understand how the ellipse is created form a keypoint
    representation. Walks through the steps I took in coming to an
    understanding.

    CommandLine:
        xdoctest ~/code/pyhesaff/tests/test_ellipse.py --show --num-samples=12

    Example:
        >>> # xdoctest: +REQUIRES(--show)
        >>> # xdoctest: +REQUIRES(module:kwimage)
        >>> # xdoctest: +REQUIRES(module:kwplot)
        >>> import kwimage
        >>> import pyhesaff
        >>> img = kwimage.grab_test_image('astro')
        >>> kpts, vecs = pyhesaff.detect_feats_in_image(img)
        >>> kp = kpts[0]
        >>> import kwplot
        >>> kwplot.autompl()
        >>> test_locals = in_depth_ellipse(kp)
        >>> kwplot.show_if_requested()
    """
    import matplotlib as mpl
    import plottool_ibeis as pt
    import vtool_ibeis.linalg as ltool
    import vtool_ibeis as vt
    import ubelt as ub
    #nSamples = 12
    nSamples = int(ub.argval('--num-samples', default=12))
    kp = np.array(kp, dtype=np.float64)

    kp = np.array([5, 6, 10.8,  3.431, 8.06, 0])
    #-----------------------
    # SETUP
    #-----------------------
    # np.set_printoptions(precision=3)
    #pt.reset()
    pt.figure(9003, docla=True, doclf=True)
    ax = pt.gca()
    ax.invert_yaxis()

    def _plotpts(data, px, color=pt.BLUE, label='', marker='.', **kwargs):
        #pt.figure(9003, docla=True, pnum=(1, 1, px))
        pt.plot2(data.T[0], data.T[1], marker, '', color=color, label=label, **kwargs)
        #pt.update()

    def _plotarrow(x, y, dx, dy, color=pt.BLUE, label=''):
        ax = pt.gca()
        arrowargs = dict(head_width=.5, length_includes_head=True, label=label)
        arrow = mpl.patches.FancyArrow(x, y, dx, dy, **arrowargs)
        arrow.set_edgecolor(color)
        arrow.set_facecolor(color)
        ax.add_patch(arrow)
        #pt.update()

    #-----------------------
    # INPUT
    #-----------------------
    print('kp = %s' % ub.repr2(kp, precision=3))
    print('--------------------------------')
    print('Let V = Perdoch.A')
    print('Let Z = Perdoch.E')
    print('Let invV = Perdoch.invA')
    print('--------------------------------')
    print("Input from Perdoch's detector: ")

    # We are given the keypoint in invA format
    if len(kp) == 5:
        (ix, iy, iv11, iv21, iv22) = kp
        iv12 = 0
    elif len(kp) == 6:
        (ix, iy, iv11, iv21, iv22, ori) = kp
        iv12 = 0
    invV = np.array([[iv11, iv12, ix],
                     [iv21, iv22, iy],
                     [   0,    0,  1]])
    V = np.linalg.inv(invV)
    Z = (V.T).dot(V)

    V_2x2 = V[0:2, 0:2]
    Z_2x2 = Z[0:2, 0:2]
    V_2x2_ = vt.decompose_Z_to_V_2x2(Z_2x2)
    assert np.all(np.isclose(V_2x2, V_2x2_))

    #C = np.linalg.cholesky(Z)
    #np.isclose(C.dot(C.T), Z)
    #Z

    print('invV is a transform from points on a unit-circle to the ellipse')
    print(ub.hzcat(['invV = ', str(invV)]))
    print('--------------------------------')
    print('V is a transformation from points on the ellipse to a unit circle')
    print(ub.hzcat(['V = ', str(V)]))
    print('--------------------------------')
    print('An ellipse is a special case of a conic. For any ellipse:')
    print('Points on the ellipse satisfy (x_ - x_0).T.dot(Z).dot(x_ - x_0) = 1')
    print('where Z = (V.T).dot(V)')
    print(ub.hzcat(['Z = ', str(Z)]))

    # Define points on a unit circle
    theta_list = np.linspace(0, TAU, nSamples)
    cicrle_pts = np.array([(np.cos(t_), np.sin(t_), 1) for t_ in theta_list])

    # Transform those points to the ellipse using invV
    ellipse_pts1 = invV.dot(cicrle_pts.T).T

    #Lets check our assertion: (x_ - x_0).T.dot(Z).dot(x_ - x_0) == 1
    x_0 = np.array([ix, iy, 1])
    checks = [(x_ - x_0).T.dot(Z).dot(x_ - x_0) for x_ in ellipse_pts1]
    try:
        # HELP: The phase is off here. in 3x3 version I'm not sure why
        #assert all([almost_eq(1, check) for check in checks1])
        is_almost_eq_pos1 = [np.allclose(1, check) for check in checks]
        is_almost_eq_neg1 = [np.allclose(-1, check) for check in checks]
        assert all(is_almost_eq_pos1)
    except AssertionError as ex:
        print('circle pts = %r ' % cicrle_pts)
        print(ex)
        print(checks)
        print([np.allclose(-1, check) for check in checks])
        raise
    else:
        #assert all([abs(1 - check) < 1E-11 for check in checks2])
        print('... all of our plotted points satisfy this')

    #=======================
    # THE CONIC SECTION
    #=======================
    # All of this was from the Perdoch paper, now lets move into conic sections
    # We will use the notation from wikipedia
    # References:
    #     http://en.wikipedia.org/wiki/Conic_section
    #     http://en.wikipedia.org/wiki/Matrix_representation_of_conic_sections

    #-----------------------
    # MATRIX REPRESENTATION
    #-----------------------
    # The matrix representation of a conic is:
    #(A,  B2, B2_, C) = Z.flatten()
    #(D, E, F) = (0, 0, 1)
    (A,  B2, D2, B2_, C, E2, D2_, E2_, F) = Z.flatten()
    B = B2 * 2
    D = D2 * 2
    E = E2 * 2
    assert B2 == B2_, 'matrix should by symmetric'
    assert D2 == D2_, 'matrix should by symmetric'
    assert E2 == E2_, 'matrix should by symmetric'
    print('--------------------------------')
    print('Now, using wikipedia\' matrix representation of a conic.')
    con = np.array((('    A', 'B / 2', 'D / 2'),
                    ('B / 2', '    C', 'E / 2'),
                    ('D / 2', 'E / 2', '    F')))
    print(ub.hzcat(['A matrix A_Q = ', str(con)]))

    # A_Q is our conic section (aka ellipse matrix)
    A_Q = np.array(((    A, B / 2, D / 2),
                    (B / 2,     C, E / 2),
                    (D / 2, E / 2,     F)))

    print(ub.hzcat(['A_Q = ', str(A_Q)]))

    #-----------------------
    # DEGENERATE CONICS
    # References:
    #    http://individual.utoronto.ca/somody/quiz.html
    print('----------------------------------')
    print('As long as det(A_Q) != it is not degenerate.')
    print('If the conic is not degenerate, we can use the 2x2 minor: A_33')
    print('det(A_Q) = %s' % str(np.linalg.det(A_Q)))
    assert np.linalg.det(A_Q) != 0, 'degenerate conic'
    A_33 = np.array(((    A, B / 2),
                     (B / 2,     C)))
    print(ub.hzcat(['A_33 = ', str(A_33)]))

    #-----------------------
    # CONIC CLASSIFICATION
    #-----------------------
    print('----------------------------------')
    print('The determinant of the minor classifies the type of conic it is')
    print('(det == 0): parabola, (det < 0): hyperbola, (det > 0): ellipse')
    print('det(A_33) = %s' % str(np.linalg.det(A_33)))
    assert np.linalg.det(A_33) > 0, 'conic is not an ellipse'
    print('... this is indeed an ellipse')

    #-----------------------
    # CONIC CENTER
    #-----------------------
    print('----------------------------------')
    print('the centers of the ellipse are obtained by: ')
    print('x_center = (B * E - (2 * C * D)) / (4 * A * C - B ** 2)')
    print('y_center = (D * B - (2 * A * E)) / (4 * A * C - B ** 2)')
    # Centers are obtained by solving for where the gradient of the quadratic
    # becomes 0. Without going through the derivation the calculation is...
    # These should be 0, 0 if we are at the origin, or our original x, y
    # coordinate specified by the keypoints. I'm doing the calculation just for
    # shits and giggles
    x_center = (B * E - (2 * C * D)) / (4 * A * C - B ** 2)
    y_center = (D * B - (2 * A * E)) / (4 * A * C - B ** 2)
    print(ub.hzcat(['x_center = ', str(x_center)]))
    print(ub.hzcat(['y_center = ', str(y_center)]))

    #-----------------------
    # MAJOR AND MINOR AXES
    #-----------------------
    # Now we are going to determine the major and minor axis
    # of this beast. It just the center augmented by the eigenvecs
    print('----------------------------------')
    # Plot ellipse axis
    # !HELP! I DO NOT KNOW WHY I HAVE TO DIVIDE, SQUARE ROOT, AND NEGATE!!!
    (evals, evecs) = np.linalg.eig(A_33)
    l1, l2 = evals
    # The major and minor axis lengths
    b = 1 / np.sqrt(l1)
    a = 1 / np.sqrt(l2)
    v1, v2 = evecs
    # Find the transformation to align the axis
    nminor = v1
    nmajor = v2
    dx1, dy1 = (v1 * b)
    dx2, dy2 = (v2 * a)
    minor = np.array([dx1, -dy1])
    major = np.array([dx2, -dy2])
    x_axis = np.array([[1], [0]])
    cosang = (x_axis.T.dot(nmajor)).T
    # Rotation angle
    theta = np.arccos(cosang)
    print('a = ' + str(a))
    print('b = ' + str(b))
    print('theta = ' + str(theta[0] / TAU) + ' * 2pi')
    # The warped eigenvects should have the same magintude
    # As the axis lengths

    try:
        assert np.allclose(a, major.dot(ltool.rotation_mat2x2(theta))[0])
        assert np.allclose(b, minor.dot(ltool.rotation_mat2x2(theta))[1])
    except AssertionError:
        print('WARNING: warped eigenvects do not have same magintude as axis length')

    try:
        # HACK
        if len(theta) == 1:
            theta = theta[0]
    except Exception:
        pass

    #-----------------------
    # ECCENTRICITY
    #-----------------------
    print('----------------------------------')
    print('The eccentricity is determined by:')
    print('')
    print('            (2 * np.sqrt((A - C) ** 2 + B ** 2))     ')
    print('ecc = -----------------------------------------------')
    print('      (nu * (A + C) + np.sqrt((A - C) ** 2 + B ** 2))')
    print('')
    print('(nu is always 1 for ellipses)')
    nu = 1
    ecc_numer = (2 * np.sqrt((A - C) ** 2 + B ** 2))
    ecc_denom = (nu * (A + C) + np.sqrt((A - C) ** 2 + B ** 2))
    ecc = np.sqrt(ecc_numer / ecc_denom)
    print('ecc = ' + str(ecc))

    # Eccentricity is a little easier in axis aligned coordinates
    # Make sure they aggree
    ecc2 = np.sqrt(1 - (b ** 2) / (a ** 2))
    try:
        assert np.allclose(ecc, ecc2), 'ecc does not aggree!'
    except Exception as ex:
        print(f'WARNING ex={ex}')

    #-----------------------
    # APPROXIMATE UNIFORM SAMPLING
    #-----------------------
    # We are given the keypoint in invA format
    print('----------------------------------')
    print('Approximate uniform points an inscribed polygon bondary')

    #def next_xy(x, y, d):
    #    # References:
    #    # http://gamedev.stackexchange.com/questions/1692/what-is-a-simple-algorithm-for-calculating-evenly-distributed-points-on-an-ellip
    #    num = (b ** 2) * (x ** 2)
    #    den = ((a ** 2) * ((a ** 2) - (x ** 2)))
    #    dxdenom = np.sqrt(1 + (num / den))
    #    deltax = d / dxdenom
    #    x_ = x + deltax
    #    y_ = b * np.sqrt(1 - (x_ ** 2) / (a ** 2))
    #    return x_, y_

    def xy_fn(t):
        return np.array((a * np.cos(t), b * np.sin(t))).T

    #nSamples = 16
    #(ix, iy, iv11, iv21, iv22), iv12 = kp, 0
    #invV = np.array([[iv11, iv12, ix],
    #                 [iv21, iv22, iy],
    #                 [   0,    0,  1]])
    #theta_list = np.linspace(0, TAU, nSamples)
    #cicrle_pts = np.array([(np.cos(t_), np.sin(t_), 1) for t_ in theta_list])
    uneven_points = invV.dot(cicrle_pts.T).T[:, 0:2]
    #uneven_points2 = xy_fn(theta_list)

    def circular_distance(arr):
        dist_most_ = ((arr[0:-1] - arr[1:]) ** 2).sum(1)
        dist_end_  = ((arr[-1] - arr[0]) ** 2).sum()
        return np.sqrt(np.hstack((dist_most_, dist_end_)))

    # Calculate the distance from each point on the ellipse to the next
    dists = circular_distance(uneven_points)
    total_dist = dists.sum()
    # Get an even step size
    multiplier = 1
    step_size = total_dist / (nSamples * multiplier)
    # Walk along edge
    num_steps_list = []
    offset_list = []
    dist_walked = 0
    total_dist = step_size
    for count in range(len(dists)):
        segment_len = dists[count]
        # Find where your starting location is
        offset_list.append(total_dist - dist_walked)
        # How far can you possibly go?
        total_dist += segment_len
        # How many steps can you take?
        num_steps = int((total_dist - dist_walked) // step_size)
        num_steps_list.append(num_steps)
        # Log how much further youve gotten
        dist_walked += (num_steps * step_size)
    #print('step_size = %r' % step_size)
    #print(np.vstack((num_steps_list, dists, offset_list)).T)

    # store the percent location at each line segment where
    # the cut will be made
    cut_list = []
    for num, dist, offset in zip(num_steps_list, dists, offset_list):
        if num == 0:
            cut_list.append([])
            continue
        offset1 = (step_size - offset) / dist
        offset2 = ((num * step_size) - offset) / dist
        cut_locs = (np.linspace(offset1, offset2, num, endpoint=True))
        cut_list.append(cut_locs)
        #print(cut_locs)

    # Cut the segments into new better segments
    approx_pts = []
    nPts = len(uneven_points)
    for count, cut_locs in enumerate(cut_list):
        for loc in cut_locs:
            pt1 = uneven_points[count]
            pt2 = uneven_points[(count + 1) % nPts]
            # Linearly interpolate between points
            new_loc = ((1 - loc) * pt1) + ((loc) * pt2)
            approx_pts.append(new_loc)
    approx_pts = np.array(approx_pts)

    # Warp approx_pts to the unit circle
    print('----------------------------------')
    print('For each aproximate point, find the closet point on the ellipse')
    #new_unit = V.dot(approx_pts.T).T
    ones_ = np.ones(len(approx_pts))
    new_hlocs = np.vstack((approx_pts.T, ones_))
    new_unit = V.dot(new_hlocs).T
    # normalize new_unit
    new_mag = np.sqrt((new_unit ** 2).sum(1))
    new_unorm_unit = new_unit / np.vstack([new_mag] * 3).T
    new_norm_unit = new_unorm_unit / np.vstack([new_unorm_unit[:, 2]] * 3).T
    # Get angle (might not be necessary)
    x_axis = np.array([1, 0, 0])
    arccos_list = x_axis.dot(new_norm_unit.T)
    uniform_theta_list = np.arccos(arccos_list)
    # Maybe this?
    uniform_theta_list = np.arctan2(new_norm_unit[:, 1], new_norm_unit[:, 0])
    #
    unevn_cicrle_pts = np.array([(np.cos(t_), np.sin(t_), 1) for t_ in uniform_theta_list])
    # This is the output. Approximately uniform points sampled along an ellipse
    uniform_ell_pts = invV.dot(unevn_cicrle_pts.T).T
    #uniform_ell_pts = invV.dot(new_norm_unit.T).T

    _plotpts(approx_pts, 0, pt.YELLOW, label='approx points', marker='o-')
    _plotpts(uniform_ell_pts, 0, pt.RED, label='uniform points', marker='o-')

    #### ALTERNATE METHOD
    import scipy.optimize
    import scipy as sp
    def angles_in_ellipse(num, a, b):
        """
        References:
            https://stackoverflow.com/questions/6972331/how-can-i-generate-a-set-of-points-evenly-distributed-along-the-perimeter-of-an
        """
        assert num > 0
        assert a < b
        angles = 2 * np.pi * np.arange(num) / num
        if a != b:
            e2 = (1.0 - a ** 2.0 / b ** 2.0)
            tot_size = sp.special.ellipeinc(2.0 * np.pi, e2)
            arc_size = tot_size / num
            arcs = np.arange(num) * arc_size
            res = sp.optimize.root(
                lambda x: (sp.special.ellipeinc(x, e2) - arcs), angles,
                options={'maxiter': 5}
            )
            angles = res.x
        return angles

    uniform_arclen_thetas = angles_in_ellipse(nSamples, a, b)
    import kwimage
    axis_aligned_optimized_uniform_pts = kwimage.Points(xy=xy_fn(uniform_arclen_thetas))
    optimized_uniform_pts = axis_aligned_optimized_uniform_pts.warp(
        kwimage.Affine.coerce(theta=theta, offset=(x_center, y_center)))

    _plotpts(optimized_uniform_pts.xy, 0, pt.GREEN, label='optimized', marker='o-')

    # Desired number of points
    #ecc = np.sqrt(1 - (b ** 2) / (a ** 2))
    # Total arclength
    #total_arclen = ellipeinc(TAU, ecc)
    #firstquad_arclen = total_arclen / 4
    # Desired arclength between points
    #d = firstquad_arclen / nSamples
    # Initial point
    #x, y = xy_fn(.001)
    #uniform_points = []
    #for count in range(nSamples):
    #    if np.isnan(x_) or np.isnan(y_):
    #        print('nan on count=%r' % count)
    #        break
    #    uniform_points.append((x_, y_))
    # The angle between the major axis and our x axis is:
    #-----------------------
    # DRAWING
    #-----------------------
    print('----------------------------------')
    # Draw the keypoint using the tried and true pt
    # Other things should subsiquently align
    #pt.draw_kpts2(np.array([kp]), ell_linewidth=4,
    #               ell_color=pt.DEEP_PINK, ell_alpha=1, arrow=True, rect=True)

    # Plot ellipse points
    _plotpts(ellipse_pts1, 0, pt.LIGHT_BLUE, label='invV.dot(cicrle_pts.T).T', marker='o-')

    _plotarrow(x_center, y_center, dx1, -dy1, color=pt.GRAY, label='minor axis')
    _plotarrow(x_center, y_center, dx2, -dy2, color=pt.GRAY, label='major axis')

    # Rotate the ellipse so it is axis aligned and plot that
    rot = ltool.rotation_around_mat3x3(theta, ix, iy)
    ellipse_pts3 = rot.dot(ellipse_pts1.T).T
    #!_plotpts(ellipse_pts3, 0, pt.GREEN, label='axis aligned points')

    # Plot ellipse orientation
    ortho_basis = np.eye(3)[:, 0:2]
    orient_axis = invV.dot(ortho_basis)
    print(orient_axis)
    _dx1, _dx2, _dy1, _dy2, _1, _2 = orient_axis.flatten()
    #!_plotarrow(x_center, y_center, _dx1, _dy1, color=pt.BLUE, label='ellipse rotation')
    #!_plotarrow(x_center, y_center, _dx2, _dy2, color=pt.BLUE)

    #pt.plt.gca().set_xlim(400, 600)
    #pt.plt.gca().set_ylim(300, 500)

    xmin, ymin = ellipse_pts1.min(0)[0:2] - 1
    xmax, ymax = ellipse_pts1.max(0)[0:2] + 1
    pt.plt.gca().set_xlim(xmin, xmax)
    pt.plt.gca().set_ylim(ymin, ymax)
    pt.legend()
    pt.dark_background(doubleit=3)
    pt.gca().invert_yaxis()

    # Hack in another view
    # It seems like the even points are not actually that even.
    # there must be a bug

    pt.figure(fnum=9003 + 1, docla=True, doclf=True, pnum=(1, 3, 1))
    _plotpts(ellipse_pts1, 0, pt.LIGHT_BLUE, label='invV.dot(cicrle_pts.T).T', marker='o-', title='even')
    pt.plt.gca().set_xlim(xmin, xmax)
    pt.plt.gca().set_ylim(ymin, ymax)
    pt.dark_background(doubleit=3)
    pt.gca().invert_yaxis()
    pt.figure(fnum=9003 + 1, pnum=(1, 3, 2))

    _plotpts(approx_pts, 0, pt.YELLOW, label='approx points', marker='o-', title='approx')
    pt.plt.gca().set_xlim(xmin, xmax)
    pt.plt.gca().set_ylim(ymin, ymax)
    pt.dark_background(doubleit=3)
    pt.gca().invert_yaxis()

    pt.figure(fnum=9003 + 1, pnum=(1, 3, 3))
    _plotpts(uniform_ell_pts, 0, pt.RED, label='uniform points', marker='o-', title='uniform')
    pt.plt.gca().set_xlim(xmin, xmax)
    pt.plt.gca().set_ylim(ymin, ymax)
    pt.dark_background(doubleit=3)
    pt.gca().invert_yaxis()

    return locals()
    # Algebraic form of connic
    #assert (a * (x ** 2)) + (b * (x * y)) + (c * (y ** 2)) + (d * x) + (e * y) + (f) == 0


#def test_ellipse_main():
#    r"""
#    CommandLine:
#        python -m pyhesaff.tests.test_ellipse --test-test_ellipse_main
#        python -m pyhesaff.tests.test_ellipse --test-test_ellipse_main --show

#    Example:
#        >>> # ENABLE_DOCTEST
#        >>> from pyhesaff.tests.test_ellipse import *  # NOQA
#        >>> # build test data
#        >>> # execute function
#        >>> result = test_ellipse_main()
#        >>> # verify results
#        >>> print(result)
#    """
#    print('__main__ = test_ellipse.py')
#    np.set_printoptions(threshold=5000, linewidth=5000, precision=3)
#    import pyhesaff.tests.pyhestest as pyhestest
#    test_data = pyhestest.load_test_data(short=True)
#    kpts = test_data['kpts']
#    kp = kpts[0]
#    #kp = np.array([0, 0, 10, 10, 10, 0])
#    print('Testing kp=%r' % (kp,))
#    test_locals = in_depth_ellipse(kp)
#    exec(ut.execstr_dict(test_locals, 'test_locals'))
#    #if '--cmd' in sys.argv:
#    #exec(helpers.execstr_dict(adaptive_locals, 'adaptive_locals'))
#    #in_depth_locals = adaptive_locals['in_depth_locals']
#    #exec(helpers.execstr_dict(in_depth_locals, 'in_depth_locals'))
#    #exec(pt.present(override1=True))
#    if ut.show_was_requested():
#        exec(pt.present())


if __name__ == '__main__':
    """
    CommandLine:
        python -m pyhesaff.tests.test_ellipse
        python -m pyhesaff.tests.test_ellipse --allexamples
        python -m pyhesaff.tests.test_ellipse --allexamples --noface --nosrc
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



# flake8: noqa
from __future__ import absolute_import, print_function, division
__version__ = '2.1.1'
__author__ = 'Jon Crall'
__author_email__ = 'erotemic@gmail.com'
__url__ = 'https://github.com/Erotemic/pyhesaff'


from pyhesaff import ctypes_interface
from pyhesaff import _pyhesaff
from pyhesaff._pyhesaff import (DESC_DIM, HESAFF_CLIB,
                                HESAFF_PARAM_DICT, HESAFF_PARAM_TYPES,
                                HESAFF_TYPED_PARAMS, KPTS_DIM,
                                adapt_scale, alloc_kpts,
                                alloc_patches, alloc_vecs,
                                argparse_hesaff_params,
                                detect_feats, detect_feats2,
                                detect_feats_in_image, detect_feats_list,
                                detect_num_feats_in_image,
                                extract_desc_from_patches, extract_patches,
                                extract_vecs, get_cpp_version,
                                get_hesaff_default_params,
                                get_is_debug_mode,
                                hesaff_kwargs_docstr_block, img32_dtype,
                                test_rot_invar,
                                vtool_adapt_rotation,)

#from pyhesaff._pyhesaff import (extract_vecs, detect_feats, detect_feats_list,
#                                adapt_scale, vtool_adapt_rotation, kpts_dtype, vecs_dtype,
#                                get_hesaff_default_params, extract_desc_from_patches,
#                                KPTS_DIM, DESC_DIM, __LIB_FPATH__, HESAFF_CLIB)
#from pyhesaff._pyhesaff import *  # NOQA
"""
# ut.util_inspect.get_module_owned_functions(pyhesaff._pyhesaff)
import pyhesaff
modname = pyhesaff.__name__
IMPORT_TUPLES = [('_pyhesaff', None)]
ut.exec_funckw(ut._internal.util_importer.dynamic_import, globals())
x = ut._internal.util_importer.dynamic_import(modname, IMPORT_TUPLES, dump=True, check_not_imported=False, verbose=True)
"""



from __future__ import absolute_import, print_function, division
import sys
from os.path import realpath, join, split
import numpy as np
import cv2
import pyhesaff


def get_test_image():
    from vtool.tests import grabdata
    img_fname = 'zebra.jpg'
    if '--zebra.png' in sys.argv:
        img_fname = 'zebra.jpg'
    if '--lena.png' in sys.argv:
        img_fname = 'lena.jpg'
    if '--jeff.png' in sys.argv:
        img_fname = 'jeff.png'
    imgdir = grabdata.get_testdata_dir()
    img_fpath = realpath(join(imgdir, img_fname))
    return img_fpath


def load_test_data(short=False, n=0, use_cpp=False, **kwargs):
    if 'short' not in vars():
        short = False
    # Read Image
    #ellipse.rrr()
    nScales = 4
    nSamples = 16
    img_fpath = get_test_image()
    imgBGR = cv2.imread(img_fpath)
    imgLAB = cv2.cvtColor(imgBGR, cv2.COLOR_BGR2LAB)
    imgL = imgLAB[:, :, 0]
    detect_kwargs = {
        'scale_min': 20,
        'scale_max': 100
    }
    detect_kwargs.update(kwargs)
    if not use_cpp:
        kpts, desc = pyhesaff.detect_feats(img_fpath, **detect_kwargs)
    else:
        # Try the new C++ code
        [kpts], [desc] = pyhesaff.detect_feats_list([img_fpath], **detect_kwargs)

    if short and n > 0:
        extra_fxs = []
        if split(img_fpath)[1] == 'zebra.png':
            extra_fxs = [374, 520, 880][0:1]
        fxs = np.array(spaced_elements2(kpts, n).tolist() + extra_fxs)
        kpts = kpts[fxs]
        desc = desc[fxs]
    test_data = locals()
    return test_data


def spaced_elements2(list_, n):
    if n is None:
        return np.arange(len(list_))
    if n == 0:
        return np.empty(0)
    indexes = np.arange(len(list_))
    stride = len(indexes) // n
    return indexes[0:-1:stride]


def spaced_elements(list_, n):
    if n is None:
        return 'list'
    indexes = np.arange(len(list_))
    stride = len(indexes) // n
    return list_[indexes[0:-1:stride]]



#!/usr/bin/env python
from __future__ import absolute_import, division, print_function
import numpy as np
import utool


def TEST_ptool_find_kpts_direction(imgBGR, kpts):
    import vtool.patch as ptool
    hrint = utool.horiz_print
    print('[rotinvar] +---')
    print('[rotinvar] | 3) Find dominant orientation in histogram')
    hrint('[rotinvar] |  * kpts.shape = ', (kpts.shape,))
    hrint('[rotinvar] |  * kpts = ', kpts)
    kpts2 = ptool.find_kpts_direction(imgBGR, kpts)
    hrint('[rotinvar] |  * kpts2.shape = ', (kpts.shape,))
    hrint('[rotinvar] |  * kpts2 = ', kpts2)
    print('[rotinvar] L___')
    return kpts2


def TEST_figure1(wpatch, gradx, grady, gmag, gori, hist, centers):
    from plottool import draw_func2 as df2
    import plottool
    import vtool.patch as ptool
    print('[rotinvar] 4) Draw histogram with interpolation annotations')
    fnum = 1
    gorimag = plottool.color_orimag(gori, gmag, True)
    nRow, nCol = (2, 7)

    df2.figure(fnum=1, pnum=(nRow, 1, nRow), doclf=True, docla=True)
    plottool.draw_hist_subbin_maxima(hist, centers)
    df2.set_xlabel('grad orientation (radians)')
    df2.set_ylabel('grad magnitude')
    df2.set_title('dominant orientations')

    print('[rotinvar] 5) Show patch, gradients, magintude, and orientation')
    df2.imshow(wpatch,    pnum=(nRow, nCol, 1), fnum=fnum, title='patch')
    df2.draw_vector_field(gradx, grady, pnum=(nRow, nCol, 2), fnum=fnum, title='gori (vec)')
    df2.imshow(gorimag, pnum=(nRow, nCol, 3), fnum=fnum, title='gori (col)')
    df2.imshow(np.abs(gradx),   pnum=(nRow, nCol, 4), fnum=fnum, title='gradx')
    df2.imshow(np.abs(grady),   pnum=(nRow, nCol, 5), fnum=fnum, title='grady')
    df2.imshow(gmag,    pnum=(nRow, nCol, 6), fnum=fnum, title='gmag')

    gpatch = ptool.gaussian_patch(shape=gori.shape)
    df2.imshow(gpatch * 255, pnum=(nRow, nCol, 7), fnum=fnum, title='gauss weights', cmap_='hot')
    #gpatch3 = np.dstack((gpatch, gpatch, gpatch))
    #df2.draw_vector_field(gradx * gpatch, grady * gpatch, pnum=(nRow, nCol, 8), fnum=fnum, title='gori (vec)')
    #df2.imshow(gorimag * gpatch3, pnum=(nRow, nCol, 9), fnum=fnum, title='gori (col)')
    #df2.imshow(gradx * gpatch,   pnum=(nRow, nCol, 10), fnum=fnum, title='gradx')
    #df2.imshow(grady * gpatch,   pnum=(nRow, nCol, 11), fnum=fnum, title='grady')
    #df2.imshow(gmag * gpatch,    pnum=(nRow, nCol, 12), fnum=fnum, title='gmag')
    return locals()


def TEST_figure2(imgBGR, kpts, desc, sel, fnum=2):
    #df2.imshow(wpatch, fnum=2)
    from plottool import draw_func2 as df2
    from plottool.viz_keypoints import _annotate_kpts, show_keypoints
    from plottool.viz_featrow import draw_feat_row
    sift = desc[sel]
    viz_kwargs = dict(ell=True, eig=False,
                      rect=True, ori_color=df2.DEEP_PINK, ell_alpha=1, fnum=fnum, pnum=(2, 1, 1))
    show_keypoints(imgBGR, kpts, sifts=None, sel_fx=sel, ori=False, **viz_kwargs)
    _annotate_kpts(kpts, sel, ori=True, **viz_kwargs)
    draw_feat_row(imgBGR, sel, kpts[sel], sift, fnum=fnum, nRows=2, nCols=3, px=3)


def TEST_keypoint(imgBGR, img_fpath, kpts, desc, sel):
    import pyhesaff
    import vtool.patch as ptool
    from plottool import draw_func2 as df2
    #----------------------#
    # --- Extract Data --- #
    #----------------------#
    kp = kpts[sel]
    # Extract patches, gradients, and orientations
    print('[rotinvar] 1) Extract patch, gradients, and orientations')
    wpatch, wkp  = ptool.get_warped_patch(imgBGR, kp, gray=True)
    gradx, grady = ptool.patch_gradient(wpatch, gaussian_weighted=False)
    gmag         = ptool.patch_mag(gradx, grady)
    gori         = ptool.patch_ori(gradx, grady)

    # Get orientation histogram
    print('[rotinvar] 2) Get orientation histogram')
    gori_weights = ptool.gaussian_weight_patch(gmag)
    hist, centers = ptool.get_orientation_histogram(gori, gori_weights)

    # Get dominant direction in radians
    kpts2 = TEST_ptool_find_kpts_direction(imgBGR, kpts)
    kpts2, desc2 = pyhesaff.vtool_adapt_rotation(img_fpath, kpts)

    #----------------------#
    # --- Draw Results --- #
    #----------------------#
    f1_loc = TEST_figure1(wpatch, gradx, grady, gmag, gori, hist, centers)
    df2.set_figtitle('Dominant Orienation Extraction')

    TEST_figure2(imgBGR, kpts, desc, sel, fnum=2)
    df2.set_figtitle('Gravity Vector')
    TEST_figure2(imgBGR, kpts2, desc2, sel, fnum=3)
    df2.set_figtitle('Rotation Invariant')

    #df2.draw_keypoint_gradient_orientations(imgBGR, kp=kpts2[sel],
    #                                        sift=desc[sel], mode='vec',
    #                                        fnum=4)

    #df2.draw_vector_field(gradx, grady, pnum=(1, 1, 1), fnum=4)
    #df2.draw_kpts2(np.array([wkp]), sifts=desc[sel:sel + 1], ori=True)
    return locals()


def test_patch_ori_main():
    r"""
    Returns:
        ?: locals_

    CommandLine:
        python -m pyhesaff.tests.test_patch_orientation --test-test_patch_ori_main
        python -m pyhesaff.tests.test_patch_orientation --test-test_patch_ori_main --show

    # Example:
    #     >>> # xdoctest: +SKIP
    #     >>> from pyhesaff.tests.test_patch_orientation import *  # NOQA
    #     >>> test_patch_ori_main()
    #     >>> import utool as ut
    #     >>> ut.show_if_requested()
    """
    import pytest
    pytest.skip('Broken in CI')

    print('[rotinvar] loading test data')
    import pyhesaff.tests.pyhestest as pyhestest
    test_data = pyhestest.load_test_data(short=True, n=3)
    img_fpath = test_data['img_fpath']
    kpts = test_data['kpts']
    desc = test_data['desc']
    imgBGR = test_data['imgBGR']
    sel = min(len(kpts) - 1, 3)

    locals_ = TEST_keypoint(imgBGR, img_fpath, kpts, desc, sel)
    return locals_

if __name__ == '__main__':
    """
    CommandLine:
        python -m pyhesaff.tests.test_patch_orientation
        python -m pyhesaff.tests.test_patch_orientation --allexamples
        python -m pyhesaff.tests.test_patch_orientation --allexamples --noface --nosrc
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals


def detect_feats_main():
    import pyhesaff
    from pyhesaff._pyhesaff import grab_test_imgpath
    from pyhesaff._pyhesaff import argparse_hesaff_params
    import cv2
    import ubelt as ub

    img_fpath = grab_test_imgpath(ub.argval('--fname', default='astro.png'))
    kwargs = argparse_hesaff_params()
    print('kwargs = %r' % (kwargs,))

    (kpts, vecs) = pyhesaff.detect_feats(img_fpath, **kwargs)

    if ub.argflag('--show'):
        # Show keypoints
        imgBGR = cv2.imread(img_fpath)
        default_showkw = dict(ori=False, ell=True, ell_linewidth=2,
                              ell_alpha=.4, ell_color='distinct')
        print('default_showkw = %r' % (default_showkw,))
        import utool as ut
        showkw = ut.argparse_dict(default_showkw)
        import plottool as pt
        pt.interact_keypoints.ishow_keypoints(imgBGR, kpts, vecs, **showkw)
        pt.show_if_requested()


def main():
    detect_feats_main()


if __name__ == '__main__':
    """
    CommandLine:
        python -m pyhesaff --show
    """
    main()



#!/usr/bin/env python
from __future__ import absolute_import, division, print_function
import numpy as np
import ubelt as ub


def TEST_ptool_find_kpts_direction(imgBGR, kpts):
    import vtool.patch as ptool
    def hrint(*x):
        print(ub.hzcat(x))
    print('[rotinvar] +---')
    print('[rotinvar] | 3) Find dominant orientation in histogram')
    hrint('[rotinvar] |  * kpts.shape = ', (kpts.shape,))
    hrint('[rotinvar] |  * kpts = ', kpts)
    kpts2 = ptool.find_kpts_direction(imgBGR, kpts)
    hrint('[rotinvar] |  * kpts2.shape = ', (kpts.shape,))
    hrint('[rotinvar] |  * kpts2 = ', kpts2)
    print('[rotinvar] L___')
    return kpts2


def TEST_figure1(wpatch, gradx, grady, gmag, gori, hist, centers, fnum=1):
    print('[rotinvar] 4) Draw histogram with interpolation annotations')
    import vtool.patch as ptool
    import plottool
    from plottool import draw_func2 as df2
    gorimag = plottool.color_orimag(gori, gmag)
    nRow, nCol = (2, 7)

    df2.figure(fnum=fnum, pnum=(nRow, 1, nRow), doclf=True, docla=True)
    plottool.draw_hist_subbin_maxima(hist, centers)
    df2.set_xlabel('grad orientation (radians)')
    df2.set_ylabel('grad magnitude')
    df2.set_title('dominant orientations')

    print('[rotinvar] 5) Show patch, gradients, magintude, and orientation')
    df2.imshow(wpatch,    pnum=(nRow, nCol, 1), fnum=fnum, title='patch')
    df2.draw_vector_field(gradx, grady, pnum=(nRow, nCol, 2), fnum=fnum, title='gori (vec)')
    df2.imshow(gorimag, pnum=(nRow, nCol, 3), fnum=fnum, title='gori (col)')
    df2.imshow(np.abs(gradx),   pnum=(nRow, nCol, 4), fnum=fnum, title='gradx')
    df2.imshow(np.abs(grady),   pnum=(nRow, nCol, 5), fnum=fnum, title='grady')
    df2.imshow(gmag,    pnum=(nRow, nCol, 6), fnum=fnum, title='gmag')

    gpatch = ptool.gaussian_patch(shape=gori.shape)
    df2.imshow(gpatch * 255, pnum=(nRow, nCol, 7), fnum=fnum, title='gauss weights', cmap_='hot')
    #gpatch3 = np.dstack((gpatch, gpatch, gpatch))
    #df2.draw_vector_field(gradx * gpatch, grady * gpatch, pnum=(nRow, nCol, 8), fnum=fnum, title='gori (vec)')
    #df2.imshow(gorimag * gpatch3, pnum=(nRow, nCol, 9), fnum=fnum, title='gori (col)')
    #df2.imshow(gradx * gpatch,   pnum=(nRow, nCol, 10), fnum=fnum, title='gradx')
    #df2.imshow(grady * gpatch,   pnum=(nRow, nCol, 11), fnum=fnum, title='grady')
    #df2.imshow(gmag * gpatch,    pnum=(nRow, nCol, 12), fnum=fnum, title='gmag')
    return locals()


def TEST_figure2(imgBGR, kpts, desc, sel, fnum=2):
    from plottool import draw_func2 as df2
    from plottool.viz_keypoints import _annotate_kpts, show_keypoints
    from plottool.viz_featrow import draw_feat_row
    #df2.imshow(wpatch, fnum=2)
    sift = desc[sel]
    viz_kwargs = dict(ell=True, eig=False,
                      rect=True, ori_color=df2.DEEP_PINK, ell_alpha=1, fnum=fnum, pnum=(2, 1, 1))
    show_keypoints(imgBGR, kpts, sifts=None, sel_fx=sel, ori=False, **viz_kwargs)
    _annotate_kpts(kpts, sel, ori=True, **viz_kwargs)
    draw_feat_row(imgBGR, sel, kpts[sel], sift, fnum=fnum, nRows=2, nCols=3, px=3)


def TEST_keypoint(imgBGR, img_fpath, kpts, desc, sel, fnum=1, figtitle=''):
    from plottool import draw_func2 as df2
    import vtool.patch as ptool
    #----------------------#
    # --- Extract Data --- #
    #----------------------#
    kp = kpts[sel]
    # Extract patches, gradients, and orientations
    print('[rotinvar] 1) Extract patch, gradients, and orientations')
    wpatch, wkp  = ptool.get_warped_patch(imgBGR, kp, gray=True)
    gradx, grady = ptool.patch_gradient(wpatch)
    gmag         = ptool.patch_mag(gradx, grady)
    gori         = ptool.patch_ori(gradx, grady)
    gori_weights = ptool.gaussian_weight_patch(gmag)

    # Get orientation histogram
    print('[rotinvar] 2) Get orientation histogram')
    hist, centers = ptool.get_orientation_histogram(gori, gori_weights)

    #----------------------#
    # --- Draw Results --- #
    #----------------------#
    f1_loc = TEST_figure1(wpatch, gradx, grady, gmag, gori, hist, centers, fnum=fnum)
    df2.set_figtitle(figtitle + 'Dominant Orienation Extraction')

    TEST_figure2(imgBGR, kpts, desc, sel, fnum=fnum + 1)
    df2.set_figtitle(figtitle)
    # TEST_figure2(imgBGR, kpts2, Desc2, sel, fnum=fnum + 2)
    # df2.set_figtitle('Rotation Invariant')

    #df2.draw_keypoint_gradient_orientations(imgBGR, kp=kpts2[sel],
    #                                        sift=desc[sel], mode='vec',
    #                                        fnum=4)

    #df2.draw_vector_field(gradx, grady, pnum=(1, 1, 1), fnum=4)
    #df2.draw_kpts2(np.array([wkp]), sifts=desc[sel:sel + 1], ori=True)
    return locals()


def test_cpp_rotinvar_main():
    r"""
    CommandLine:
        python -m pyhesaff.tests.test_cpp_rotation_invariance --test-test_cpp_rotinvar_main
        python -m pyhesaff.tests.test_cpp_rotation_invariance --test-test_cpp_rotinvar_main --show

    # Example:
    #     >>> # DISABLE_DOCTEST
    #     >>> from pyhesaff.tests.test_cpp_rotation_invariance import *  # NOQA
    #     >>> # build test data
    #     >>> # execute function
    #     >>> result = test_cpp_rotinvar_main()
    #     >>> # verify results
    #     >>> print(result)
    """
    import pytest
    pytest.skip('Broken in CI')
    # TODO; take visualization out of this test by default
    from pyhesaff.tests import pyhestest
    import pyhesaff
    # Read data
    print('[rotinvar] loading test data')
    img_fpath = pyhestest.get_test_image()
    [kpts1], [desc1] = pyhesaff.detect_feats_list([img_fpath], rotation_invariance=False)
    [kpts2], [desc2] = pyhesaff.detect_feats_list([img_fpath], rotation_invariance=True)
    np.set_printoptions(threshold=5000, linewidth=5000, precision=8, suppress=True)

    print('kpts1.shape = %r' % (kpts1.shape,))
    print('kpts2.shape = %r' % (kpts2.shape,))

    print('desc1.shape = %r' % (desc1.shape,))
    print('desc2.shape = %r' % (desc2.shape,))

    print('\n----\n'.join([str(k1) + '\n' + str(k2) for k1, k2 in zip(kpts1[0:10], kpts2[0:10])]))

    n = 4
    #clip = min(len(kpts1), n)

    # HACK FIXME
    fxs = np.array(pyhestest.spaced_elements2(kpts2, n).tolist()[0:3])

    print('fxs=%r' % fxs)
    kpts1 = kpts1[fxs]
    kpts2 = kpts2[fxs]
    desc1 = desc1[fxs]
    desc2 = desc2[fxs]

    print('\n----\n'.join([str(k1) + '\n' + str(k2) for k1, k2 in zip(kpts1, kpts2)]))

    imgBGR = pyhestest.cv2.imread(img_fpath)
    sel = min(len(kpts1) - 1, 3)

    TEST_keypoint(imgBGR, img_fpath, kpts1, desc1, sel, fnum=1, figtitle='Downward Rotation')
    TEST_keypoint(imgBGR, img_fpath, kpts2, desc2, sel, fnum=9001, figtitle='Adapted Rotation')

    #locals_ = TEST_keypoint(imgBGR, img_fpath, kpts1, desc1, sel)

    #pinteract.interact_keypoints(imgBGR, kpts2, desc, arrow=True, rect=True)
    if ub.argflag('--show'):
        from plottool import draw_func2 as df2
        exec(df2.present())


if __name__ == '__main__':
    """
    CommandLine:
        python ~/code/pyhesaff/tests/test_cpp_rotation_invariance.py
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



#!/usr/bin/env python
from __future__ import absolute_import, division, print_function
from six.moves import zip
import numpy as np


def do_test_hesaff_kpts(img_fpath, **kwargs):
    import pytest
    pytest.skip('Broken in CI')
    if 'kwargs' not in vars():
        kwargs = {}
    # Make detector and read image
    import pyhesaff
    hesaff_ptr = pyhesaff.new_hesaff(img_fpath, **kwargs)
    # Return the number of keypoints detected
    nKpts = pyhesaff.hesaff_lib.detect(hesaff_ptr)
    #print('[pyhesaff] detected: %r keypoints' % nKpts)
    # Allocate arrays
    kpts = np.empty((nKpts, pyhesaff.KPTS_DIM), pyhesaff.kpts_dtype)
    desc = np.empty((nKpts, pyhesaff.DESC_DIM), pyhesaff.desc_dtype)
    # Populate arrays
    pyhesaff.hesaff_lib.exportArrays(hesaff_ptr, nKpts, kpts, desc)
    # TODO: Incorporate parameters
    # TODO: Scale Factor
    #hesaff_lib.extractDesc(hesaff_ptr, nKpts, kpts, desc2)
    #hesaff_lib.extractDesc(hesaff_ptr, nKpts, kpts, desc3)
    #print('[hesafflib] returned')
    return kpts, desc


def test_adaptive_scale():
    # Get relevant test data
    """
    __name__ = 'IPython'
    exec(open('test_pyhesaff.py').read())
    exec(open('vtellipse.py').read())
    """
    import pytest
    pytest.skip('Broken in CI')
    print('test_adaptive_scale()')
    import vtool.ellipse as vtellipse
    from plottool import draw_func2 as df2
    from plottool.viz_keypoints import show_keypoints
    from pyhesaff.tests import pyhestest
    test_data = pyhestest.load_test_data(short=True, n=4)
    img_fpath = test_data['img_fpath']
    imgL = test_data['imgL']
    imgBGR = test_data['imgBGR']
    kpts = test_data['kpts']
    desc = test_data['desc']
    # WHY ARENT THESE 1s?
    #np.sqrt(((desc / 256.0) ** 2).sum(1))

    nScales = 16
    nSamples = 16
    low, high = -1, 2
    nKp = len(kpts)

    df2.figure(fnum=1, doclf=True, docla=True)

    def show_kpts(kpts_, px, title):
        show_keypoints(imgBGR, kpts_, pnum=(2, 3, px + 3), fnum=1,
                       color=df2.BLUE, title=title)

    def plot_line(vals, title):
        df2.figure(fnum=1, pnum=(2, 1, 1))
        df2.plot(vals, 'bo-')

    def plot_marks(vals, marker, title):
        df2.figure(fnum=1, pnum=(2, 1, 1))
        df2.plot(vals[0], vals[1], marker)

    fx = 0

    # INPUT
    kpts = np.array(kpts, dtype=np.float64)
    show_kpts(kpts, 1, 'original keypoints')

    # STEP1: EXPAND KEYPOINTS
    print('step1: expand_keypoints()')
    expanded_kpts = vtellipse.expand_scales(kpts, nScales, low, high)
    show_kpts(expanded_kpts, 2, 'expanded keypoint')

    # STEP2: UNIFORM SAMPLE / INTERPOLATE MAXIMA
    print('step2: uniform_sample / interpolate maxima()')
    border_vals_sum = vtellipse.sample_ell_border_vals(imgBGR, expanded_kpts, nKp, nScales, nSamples)
    x_data_list, y_data_list = vtellipse.find_maxima_with_neighbors(border_vals_sum)
    peak_list = vtellipse.interpolate_peaks(x_data_list, y_data_list)

    plot_line(border_vals_sum[fx], 'gradient mag')
    plot_marks([x_data_list[fx][0], y_data_list[fx][0]], 'go', 'left')
    plot_marks([x_data_list[fx][1], y_data_list[fx][1]], 'ro', 'extreme')
    plot_marks([x_data_list[fx][2], y_data_list[fx][2]], 'go', 'right')
    plot_marks(peak_list[fx].T, 'rx', 'interpolated peaks')

    parabolas_list = []
    for x_data, y_data in zip(x_data_list, y_data_list):
        parabola_points = []
        for (x1, x2, x3), (y1, y2, y3) in zip(x_data.T, y_data.T):
            coeff = np.polyfit((x1, x2, x3), (y1, y2, y3), 2)
            xpoints = np.linspace(x1, x3, 50)
            ypoints = np.polyval(coeff, xpoints)
            parabola_points.append((xpoints, ypoints))
        parabolas_list.append(parabola_points)

    df2.set_ylabel('gradient magnitude')
    df2.set_xlabel('scale index')

    for xdat, ydat in parabolas_list[0]:
        plot_marks([xdat, ydat], 'g--', '')

    # STEP 3: INTERPOLATE SCALES
    print('step3: interpolate scales')
    subscale_list = vtellipse.interpolate_between(peak_list, nScales, high, low)
    subscale_kpts = vtellipse.expand_subscales(kpts, subscale_list)
    #show_kpts(subscale_kpts, 3, 'subscale keypoint')

    # STEP 4: Check Image Bounds
    print('step4: check image bounds')
    # Make sure that the new shapes are in bounds
    height, width = imgBGR.shape[0:2]
    isvalid = vtellipse.check_kpts_in_bounds(subscale_kpts, width, height)
    adapted_kpts = np.array(subscale_kpts[isvalid], dtype=np.float32)
    show_kpts(adapted_kpts, 3, 'adapted keypoint')

    #df2.update()

    scales = 2 ** np.linspace(low, high, nScales)
    adapted_kpts = vtellipse.adaptive_scale(img_fpath, kpts, nScales, low, high, nSamples)

    #plot_vals(adapted_kpts, pnum=(3

    #viz.show_keypoints(imgBGR, adapted_kpts, pnum=(3, 1, 3), fnum=1, color=df2.BLUE, title='adapted keypoints')

    #adapted_desc = pyhesaff.extract_desc(img_fpath, adapted_kpts)
    #desc = pyhesaff.extract_desc(img_fpath, kpts)
    ##interact.interact_keypoints(imgBGR, adapted_kpts, adapted_desc)
    #df2.update()
    return locals()


def test_adaptive_scale_main():
    r"""
    CommandLine:
        python -m pyhesaff.tests.test_adaptive_scale --test-test_adaptive_scale_main
        python -m pyhesaff.tests.test_adaptive_scale --test-test_adaptive_scale_main --show

    # Example:
    #     >>> # DISABLE_DOCTEST
    #     >>> from pyhesaff.tests.test_adaptive_scale import *  # NOQA
    #     >>> # build test data
    #     >>> # execute function
    #     >>> result = test_adaptive_scale_main()
    #     >>> # verify results
    #     >>> print(result)
    """
    import pytest
    pytest.skip('Broken in CI')
    print('__main__ = test_adaptive_scale.py')
    from plottool import draw_func2 as df2
    import ubelt as ub
    np.set_printoptions(threshold=5000, linewidth=5000, precision=3)

    #adaptive_locals = test_adaptive_scale()
    # They seem to work
    # TODO: take the gui functions out of this test
    test_adaptive_scale()

    if ub.argflag('--show'):
        exec(df2.present())


if __name__ == '__main__':
    """
    CommandLine:
        python -m pyhesaff.tests.test_adaptive_scale
        python -m pyhesaff.tests.test_adaptive_scale --allexamples
        python -m pyhesaff.tests.test_adaptive_scale --allexamples --noface --nosrc
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



"""
This helps find the shared library that contains the compiled subroutines.
Its a bit hacky and could use a cleanup by someone who really understands
how python c-extension libraries are named and placed depending on system.
"""
from os.path import join, exists, normpath
import sys
import os
import ctypes


#============================
# general ctypes interface
#============================

__DEBUG_CLIB__ = '--debug' in sys.argv or '--debug-clib' in sys.argv


def get_plat_specifier():
    """
    Standard platform specifier used by distutils
    """
    import distutils
    try:
        plat_name = distutils.util.get_platform()
    except AttributeError:
        plat_name = distutils.sys.platform
    plat_specifier = '.{}-{}'.format(plat_name, sys.version[0:3])
    if hasattr(sys, 'gettotalrefcount'):
        plat_specifier += '-pydebug'
    return plat_specifier


def get_candidate_plat_specifiers():
    import distutils
    if sys.maxsize > 2 ** 32:
        arch = 'x86_64'  # TODO: get correct arch spec
    else:
        arch = 'i686'  # TODO: get correct arch spec

    py_ver = sys.version[0:3]

    try:
        plat_name = distutils.util.get_platform()
    except AttributeError:
        plat_name = distutils.sys.platform

    plat_name_cands = [plat_name]
    if sys.platform.startswith('linux'):
        plat_name_cands.append('linux')
        plat_name_cands.append('manylinux1')
        plat_name_cands.append('manylinux')
    elif sys.platform.startswith('darwin'):
        # HACK:
        # on travis, wheel builds as libhesaff.macosx-10.12-x86_64-2.7.dylib,
        # but we seem to want libhesaff.macosx-10.6-intel-2.7.dylib
        # TODO: what is the proper way to determine the ABI tag?
        plat_name_cands.append('macosx-10.6')
        plat_name_cands.append('macosx-10.7')
        plat_name_cands.append('macosx-10.9')
        plat_name_cands.append('macosx-10.12')
        plat_name_cands.append('macosx-10.6-intel')
        plat_name_cands.append('macosx-10.7-intel')
        plat_name_cands.append('macosx-10.9-intel')
        plat_name_cands.append('macosx-10.12-intel')
    elif sys.platform.startswith('win32'):
        # hack for win32
        plat_name_cands.append('win-amd64')
        pass

    spec_list = []
    for plat_name in plat_name_cands:
        spec_list.extend([
            '.{}-{}'.format(plat_name, sys.version[0:3]),
            '.{}-{}-{}'.format(plat_name, arch, py_ver),
        ])
    spec_list.append('')
    return spec_list


def get_lib_fname_candidates(libname):
    """
    Args:
        libname (str): library name (e.g. 'hesaff', not 'libhesaff')

    Returns:
        list: libnames - list of plausible library file names

    CommandLine:
        python -m pyhesaff.ctypes_interface get_lib_fname_candidates

    Example:
        >>> from pyhesaff.ctypes_interface import *  # NOQA
        >>> libname = 'hesaff'
        >>> libnames = get_lib_fname_candidates(libname)
        >>> import ubelt as ub
        >>> print('libnames = {}'.format(ub.repr2(libnames)))
    """
    spec_list = get_candidate_plat_specifiers()

    prefix_list = ['lib' + libname]
    if sys.platform.startswith('win32'):
        # windows doesnt start names with lib
        prefix_list.append(libname)
        ext = '.dll'
    elif sys.platform.startswith('darwin'):
        ext = '.dylib'
    elif sys.platform.startswith('linux'):
        ext = '.so'
    else:
        raise Exception('Unknown operating system: %s' % sys.platform)
    # Construct priority ordering of libnames
    libnames = [''.join((prefix, spec, ext))
                for spec in spec_list
                for prefix in prefix_list]
    return libnames


def get_lib_dpath_list(root_dir):
    """
    input <root_dir>: deepest directory to look for a library (dll, so, dylib)
    returns <libnames>: list of plausible directories to look.
    """
    'returns possible lib locations'
    get_lib_dpath_list = [
        root_dir,
        # join(root_dir, 'lib'),
        # join(root_dir, 'build'),
        # join(root_dir, 'build', 'lib'),
    ]
    return get_lib_dpath_list


def find_lib_fpath(libname, root_dir, verbose=False):
    """ Search for the library """
    lib_fname_list = get_lib_fname_candidates(libname)
    tried_fpaths = []

    FINAL_LIB_FPATH = None

    for lib_fname in lib_fname_list:
        if verbose:
            print('--')
        curr_dpath = root_dir
        # max_depth = 0

        for lib_dpath in get_lib_dpath_list(curr_dpath):
            lib_fpath = normpath(join(lib_dpath, lib_fname))
            tried_fpaths.append(lib_fpath)
            flag = exists(lib_fpath)
            if verbose:
                print('[c] Check: {}, exists={}'.format(lib_fpath, int(flag)))
            if flag:
                if verbose:
                    print('using: {}'.format(lib_fpath))
                FINAL_LIB_FPATH = lib_fpath
                return lib_fpath

    if FINAL_LIB_FPATH is not None:
        return FINAL_LIB_FPATH
    else:
        contents = os.listdir(root_dir)
        msg = ('\n[C!] find_lib_fpath(libname={!r}, root_dir={!r})'.format(
               libname, root_dir) +
               '\n[c!] Cannot FIND dynamic library')
        print(msg)
        print('\n[c!] Checked: '.join(tried_fpaths))
        print('UNABLE TO FIND LIB IN DPATH contents = {!r}'.format(contents))
        raise ImportError(msg)


def load_clib(libname, root_dir):
    """
    Searches for a library matching libname and loads it

    Args:
        libname:  library name (e.g. 'hesaff', not 'libhesaff')

        root_dir: the directory that should contain the
                  library file (dll, dylib, or so).
    Returns:
        clib: a ctypes object used to interface with the library
    """
    ex = None
    lib_fpath = find_lib_fpath(libname, root_dir)
    try:
        if sys.platform.startswith('win32'):
            clib = ctypes.windll[lib_fpath]
        else:
            clib = ctypes.cdll[lib_fpath]
    except OSError as ex_:
        ex = ex_
        print('[C!] Caught OSError:\n{!r}'.format(ex))
        errsuffix = 'Is there a missing dependency?'
    except Exception as ex_:
        ex = ex_
        print('[C!] Caught Exception:\n{!r}'.format(ex))
        errsuffix = 'Was the library correctly compiled?'
    else:
        def def_cfunc(return_type, func_name, arg_type_list):
            'Function to define the types that python needs to talk to c'
            cfunc = getattr(clib, func_name)
            cfunc.restype = return_type
            cfunc.argtypes = arg_type_list
        clib.__LIB_FPATH__ = lib_fpath
        return clib, def_cfunc, lib_fpath
    print('[C!] cwd={!r}'.format(os.getcwd()))
    print('[C!] load_clib(libname={!r}, root_dir={!r})'.format(libname, root_dir))
    print('[C!] lib_fpath = {!r}'.format(lib_fpath))
    errmsg = '[C] Cannot LOAD {!r} dynamic library. Caused by ex={!r}. {}'.format(libname, ex, errsuffix)
    print(errmsg)
    raise ImportError(errmsg)


if __name__ == '__main__':
    r"""
    CommandLine:
        python -m pyhesaff.ctypes_interface
        python -m pyhesaff.ctypes_interface --allexamples
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



#!/usr/bin/env python
from __future__ import absolute_import, division, print_function
import cv2
import utool as ut


def interactive_test_pyheaff(img_fpath):
    r"""
    This show is interactive in this test!

    Args:
        img_fpath (str):

    # Example:
    #     >>> # ENABLE_DOCTEST
    #     >>> from pyhesaff.tests.test_pyhesaff import *  # NOQA
    #     >>> img_fpath = ut.grab_test_imgpath('jeff.png')
    #     >>> interactive_test_pyheaff(img_fpath)
    """
    import pytest
    pytest.skip('Broken in CI')

    import pyhesaff
    kpts, desc = pyhesaff.detect_feats(img_fpath)
    rchip = cv2.imread(img_fpath)
    if ut.show_was_requested():
        from plottool.interact_keypoints import ishow_keypoints
        ishow_keypoints(rchip, kpts, desc)
    return locals()


if __name__ == '__main__':
    import xdoctest
    xdoctest.doctest_module(__file__)



#!/usr/bin/env python
from __future__ import absolute_import, division, print_function
import numpy as np
from six.moves import zip
import ubelt as ub
import utool as ut


def test_simple_parallel():
    r"""
    CommandLine:
        python ~/code/pyhesaff/tests/test_pyhesaff_simple_parallel.py

    # Example:
    #     >>> # ENABLE_DOCTEST
    #     >>> from pyhesaff.tests.test_pyhesaff_simple_parallel import *  # NOQA
    #     >>> import matplotlib as mpl
    #     >>> from matplotlib import pyplot as plt
    #     >>> img_fpaths, kpts_array, desc_array = test_simple_parallel()
    #     >>> ut.quit_if_noshow()
    #     >>> # Do not plot by default
    #     >>> fig = plt.figure()
    #     >>> for count, (img_fpath, kpts, desc) in enumerate(zip(img_fpaths, kpts_array,
    #     >>>                                                     desc_array)):
    #     >>>     if count > 3:
    #     >>>         break
    #     >>>     ax = fig.add_subplot(2, 2, count + 1)
    #     >>>     img = mpl.image.imread(img_fpath)
    #     >>>     plt.imshow(img)
    #     >>>     _xs, _ys = kpts.T[0:2]
    #     >>>     ax.plot(_xs, _ys, 'ro', alpha=.5)
    #     >>> ut.show_if_requested()
    """
    import pytest
    pytest.skip('Broken in CI')

    import pyhesaff
    # import kwimage
    test_fnames = ['carl.jpg', 'lena.png', 'zebra.png', 'ada.jpg', 'star.png']
    img_fpaths = list(map(ut.grab_test_imgpath, test_fnames)) * 2

    # Time parallel computation
    with ub.Timer('Timing Parallel'):
        kpts_array, desc_array = pyhesaff.detect_feats_list(img_fpaths)

    # Time serial computation
    kpts_list2 = []
    desc_list2 = []
    with ub.Timer('Timing Iterative'):
        for img_fpath in img_fpaths:
            kpts_, desc_ = pyhesaff.detect_feats(img_fpath)
            kpts_list2.append(kpts_)
            desc_list2.append(desc_)

    print('Checking for errors')
    for (kpts_, desc_, kpts, desc) in zip(kpts_list2, desc_list2, kpts_array, desc_array):
        print('shape(kpts, kpts_, desc, desc_) = %9r, %9r, %11r, %11r' %
              (kpts.shape, kpts_.shape, desc.shape, desc_.shape))
        try:
            assert np.all(kpts_ == kpts), 'parallel computation inconsistent'
            assert np.all(desc_ == desc), 'parallel computation inconsistent'
            assert len(kpts_) > 0, 'no kpts detected'
            #assert False, 'deliberate triggering to see printouts'
        except Exception:
            raise
    print('Keypoints seem consistent')
    return img_fpaths, kpts_array, desc_array


if __name__ == '__main__':
    """
    CommandLine:
        python ~/code/pyhesaff/tests/test_pyhesaff_simple_parallel.py
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



#!/usr/bin/env python
from __future__ import absolute_import, division, print_function
import utool as ut


def test_simple_iterative():
    r"""
    CommandLine:
        python -m pyhesaff.tests.test_pyhesaff_simple_iterative --test-simple_iterative_test
        python -m pyhesaff.tests.test_pyhesaff_simple_iterative --test-simple_iterative_test --show
    """
    import pytest
    pytest.skip('Broken in CI')

    import pyhesaff
    fpath_list = [
        ut.grab_test_imgpath('lena.png'),
        ut.grab_test_imgpath('carl.jpg'),
        ut.grab_test_imgpath('grace.jpg'),
        ut.grab_test_imgpath('ada.jpg'),
    ]
    kpts_list = []

    for img_fpath in fpath_list:
        kpts, vecs = pyhesaff.detect_feats(img_fpath)
        print('img_fpath=%r' % img_fpath)
        print('kpts=%s' % (ut.truncate_str(repr(kpts)),))
        print('vecs=%s' % (ut.truncate_str(repr(vecs)),))
        assert len(kpts) == len(vecs)
        assert len(kpts) > 0, 'no keypoints were detected!'
        kpts_list.append(kpts)

    if ut.show_was_requested():
        import matplotlib as mpl
        from matplotlib import pyplot as plt
        fig = plt.figure()
        for i, fpath, kpts in enumerate(zip(fpath_list, kpts_list), start=1):
            ax = fig.add_subplot(2, 2, i)
            img = mpl.image.imread(fpath)
            plt.imshow(img)
            _xs, _ys = kpts.T[0:2]
            ax.plot(_xs, _ys, 'ro', alpha=.5)


if __name__ == '__main__':
    """
    CommandLine:
        python -m pyhesaff.tests.test_pyhesaff_simple_iterative simple_iterative_test
        python -m pyhesaff.tests.test_pyhesaff_simple_iterative --allexamples
        python -m pyhesaff.tests.test_pyhesaff_simple_iterative --allexamples --noface --nosrc
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



#!/usr/bin/env python
"""
The python hessian affine keypoint module

Command Line:
    python -m pyhesaff detect_feats --show --siftPower=0.5 --maxBinValue=-1
    python -m pyhesaff detect_feats --show
    python -m pyhesaff detect_feats --show --siftPower=0.5,
"""
import ctypes
import numpy as np
import ubelt as ub
from os.path import realpath, dirname
from collections import OrderedDict
from pyhesaff import ctypes_interface

#============================
# hesaff ctypes interface
#============================

# numpy dtypes
kpts_dtype = np.float32
vecs_dtype = np.uint8
img_dtype  = np.uint8
img32_dtype  = np.float32
# scalar ctypes
obj_t     = ctypes.c_void_p
str_t     = ctypes.c_char_p
int_t     = ctypes.c_int
#else:
#    raise NotImplementedError('PY3')
#    int_t     = ctypes.c_long
bool_t    = ctypes.c_bool
float_t   = ctypes.c_float
#byte_t    = ctypes.c_char
# array ctypes
FLAGS_RW = str('aligned, c_contiguous, writeable')
FLAGS_RO = str('aligned, c_contiguous')
#FLAGS_RW = 'aligned, writeable'
kpts_t       = np.ctypeslib.ndpointer(dtype=kpts_dtype, ndim=2, flags=FLAGS_RW)
vecs_t       = np.ctypeslib.ndpointer(dtype=vecs_dtype, ndim=2, flags=FLAGS_RW)
img_t        = np.ctypeslib.ndpointer(dtype=img_dtype, ndim=3, flags=FLAGS_RO)
img32_t      = np.ctypeslib.ndpointer(dtype=img32_dtype, ndim=3, flags=FLAGS_RO)
kpts_array_t = np.ctypeslib.ndpointer(dtype=kpts_t, ndim=1, flags=FLAGS_RW)
vecs_array_t = np.ctypeslib.ndpointer(dtype=vecs_t, ndim=1, flags=FLAGS_RW)
int_array_t  = np.ctypeslib.ndpointer(dtype=int_t, ndim=1, flags=FLAGS_RW)
str_list_t   = ctypes.POINTER(str_t)

# THE ORDER OF THIS LIST IS IMPORTANT!
HESAFF_TYPED_PARAMS = [
    # Pyramid Params
    (int_t,   'numberOfScales', 3),           # number of scale per octave
    (float_t, 'threshold', 16.0 / 3.0),       # noise dependent threshold on the response (sensitivity)
    (float_t, 'edgeEigenValueRatio', 10.0),   # ratio of the eigenvalues
    (int_t,   'border', 5),                   # number of pixels ignored at the border of image
    (int_t,   'maxPyramidLevels', -1),        # maximum number of pyramid divisions. -1 is no limit
    # Affine Shape Params
    (int_t,   'maxIterations', 16),           # number of affine shape interations
    (float_t, 'convergenceThreshold', 0.05),  # maximum deviation from isotropic shape at convergence
    (int_t,   'smmWindowSize', 19),           # width and height of the SMM (second moment matrix) mask
    (float_t, 'mrSize', 3.0 * np.sqrt(3.0)),  # size of the measurement region (as multiple of the feature scale)
    # SIFT params
    (int_t,   'spatialBins', 4),
    (int_t,   'orientationBins', 8),
    (float_t, 'maxBinValue', 0.2),
    # Shared params
    (float_t, 'initialSigma', 1.6),           # amount of smoothing applied to the initial level of first octave
    (int_t,   'patchSize', 41),               # width and height of the patch
    # My params
    (float_t, 'scale_min', -1.0),
    (float_t, 'scale_max', -1.0),
    (bool_t,  'rotation_invariance', False),
    (bool_t,  'augment_orientation', False),
    (float_t, 'ori_maxima_thresh', .8),
    (bool_t,  'affine_invariance', True),
    (bool_t,  'only_count', False),
    #
    (bool_t,  'use_dense', False),
    (int_t,   'dense_stride', 32),
    (float_t, 'siftPower', 1.0),
]

HESAFF_PARAM_DICT = OrderedDict([(key, val) for (type_, key, val) in HESAFF_TYPED_PARAMS])
HESAFF_PARAM_TYPES = [type_ for (type_, key, val) in HESAFF_TYPED_PARAMS]


def grab_test_imgpath(p):
    fpath = ub.grabdata('https://i.imgur.com/KXhKM72.png',
                        fname='astro.png',
                        hash_prefix='160b6e5989d2788c0296eac45b33e90fe612da23',
                        hasher='sha1')
    return fpath


def imread(fpath):
    import cv2
    return cv2.imread(fpath)


def _build_typed_params_kwargs_docstr_block(typed_params):
    r"""
    Args:
        typed_params (dict):

    CommandLine:
        python -m pyhesaff build_typed_params_docstr

    Example:
        >>> # DISABLE_DOCTEST
        >>> from pyhesaff._pyhesaff import *  # NOQA
        >>> typed_params = HESAFF_TYPED_PARAMS
        >>> result = build_typed_params_docstr(typed_params)
        >>> print(result)
    """
    kwargs_lines = []
    for tup in typed_params:
        type_, name, default = tup
        typestr = str(type_).replace('<class \'ctypes.c_', '').replace('\'>', '')
        line_fmtstr = '{name} ({typestr}): default={default}'
        line = line_fmtstr.format(name=name, typestr=typestr, default=default)
        kwargs_lines.append(line)
    kwargs_docstr_block = ('Kwargs:\n' + ub.indent('\n'.join(kwargs_lines), '    '))
    return ub.indent(kwargs_docstr_block, '    ')
hesaff_kwargs_docstr_block = _build_typed_params_kwargs_docstr_block(HESAFF_TYPED_PARAMS)


HESAFF_CLIB = None


def argparse_hesaff_params():
    alias_dict = {'affine_invariance': 'ai'}
    alias_dict = {'rotation_invariance': 'ri'}
    default_dict_ = get_hesaff_default_params()
    try:
        import utool as ut
        hesskw = ut.argparse_dict(default_dict_, alias_dict=alias_dict)
    except Exception as ex:
        print('ex = {!r}'.format(ex))
        return default_dict_
    return hesskw


def _load_hesaff_clib():
    """
    Specificially loads the hesaff lib and defines its functions
    """
    # Get the root directory which should have the dynamic library in it
    #root_dir = realpath(dirname(__file__)) if '__file__' in vars() else realpath(os.getcwd())

    # os.path.dirname(sys.executable)
    #if getattr(sys, 'frozen', False):
    #    # we are running in a |PyInstaller| bundle
    #     root_dir = realpath(sys._MEIPASS)
    #else:
    #    # we are running in a normal Python environment
    #    root_dir = realpath(dirname(__file__))
    root_dir = realpath(dirname(__file__))

    libname = 'hesaff'
    (clib, def_cfunc, lib_fpath) = ctypes_interface.load_clib(libname, root_dir)
    # Expose extern C Functions to hesaff's clib
    #def_cfunc(ctypes.c_char_p, 'cmake_build_type',       [])
    #def_cfunc(None,  'free_char',       [ctypes.c_char_p])
    def_cfunc(int_t, 'get_cpp_version',        [])
    def_cfunc(int_t, 'is_debug_mode',          [])
    def_cfunc(int_t, 'detect',                 [obj_t])
    def_cfunc(int_t, 'get_kpts_dim',           [])
    def_cfunc(int_t, 'get_desc_dim',           [])
    def_cfunc(None,  'exportArrays',           [obj_t, int_t, kpts_t, vecs_t])
    def_cfunc(None,  'extractDesc',            [obj_t, int_t, kpts_t, vecs_t])
    def_cfunc(None,  'extractPatches',         [obj_t, int_t, kpts_t, img32_t])
    def_cfunc(None,  'extractDescFromPatches', [int_t, int_t, int_t, img_t, vecs_t])
    def_cfunc(obj_t, 'new_hesaff_fpath',       [str_t] + HESAFF_PARAM_TYPES)
    def_cfunc(obj_t, 'new_hesaff_image',       [img_t, int_t, int_t, int_t] + HESAFF_PARAM_TYPES)
    def_cfunc(None,  'free_hesaff',            [obj_t])
    def_cfunc(obj_t, 'detectFeaturesListStep1',   [int_t, str_list_t] + HESAFF_PARAM_TYPES)
    def_cfunc(None,  'detectFeaturesListStep2',    [int_t, obj_t, int_array_t])
    def_cfunc(None,  'detectFeaturesListStep3',    [int_t, obj_t, int_array_t, int_array_t, kpts_t, vecs_t])
    return clib, lib_fpath

# Create a global interface to the hesaff lib
try:
    HESAFF_CLIB, __LIB_FPATH__ = _load_hesaff_clib()
except AttributeError:
    print('Need to rebuild hesaff')
    raise

KPTS_DIM = HESAFF_CLIB.get_kpts_dim()
DESC_DIM = HESAFF_CLIB.get_desc_dim()


#============================
# helpers
#============================


def alloc_patches(nKpts, size=41):
    patches = np.empty((nKpts, size, size), np.float32)
    return patches


def alloc_vecs(nKpts):
    # array of bytes
    vecs = np.empty((nKpts, DESC_DIM), vecs_dtype)
    return vecs


def alloc_kpts(nKpts):
    # array of floats
    kpts = np.empty((nKpts, KPTS_DIM), kpts_dtype)
    #kpts = np.zeros((nKpts, KPTS_DIM), kpts_dtype) - 1.0  # array of floats
    return kpts


def _make_hesaff_cpp_params(kwargs):
    hesaff_params = HESAFF_PARAM_DICT.copy()
    for key, val in kwargs.items():
        if key in hesaff_params:
            hesaff_params[key] = val
        else:
            print('[pyhesaff] WARNING: key=%r is not known' % key)
    return hesaff_params


def _cast_strlist_to_C(py_strlist):
    """
    Converts a python list of strings into a c array of strings

    References:
        http://stackoverflow.com/questions/3494598/pass-list-of-strings-ctypes
    """
    c_strarr = (str_t * len(py_strlist))()
    c_strarr[:] = py_strlist
    return c_strarr


def _new_fpath_hesaff(img_fpath, **kwargs):
    """ Creates new detector object which reads the image """
    hesaff_params = _make_hesaff_cpp_params(kwargs)
    hesaff_args = hesaff_params.values()  # pass all parameters to HESAFF_CLIB
    img_realpath = realpath(img_fpath)
    # convert out of unicode
    img_realpath = img_realpath.encode('ascii')
    try:
        hesaff_ptr = HESAFF_CLIB.new_hesaff_fpath(img_realpath, *hesaff_args)
    except Exception:
        msg = 'hesaff_ptr = HESAFF_CLIB.new_hesaff_fpath(img_realpath, *hesaff_args)',
        print('msg = {!r}'.format(msg))
        print('hesaff_args = {!r}'.format(hesaff_args))
        raise
    return hesaff_ptr


def _new_image_hesaff(img, **kwargs):
    """ Creates new detector object which reads the image """
    hesaff_params = _make_hesaff_cpp_params(kwargs)
    hesaff_args = hesaff_params.values()  # pass all parameters to HESAFF_CLIB
    rows, cols = img.shape[0:2]
    if len(img.shape) == 2:
        channels = 1
    else:
        channels = img.shape[2]
    try:
        hesaff_ptr = HESAFF_CLIB.new_hesaff_image(
            img, rows, cols, channels, *hesaff_args)
    except Exception:
        msg = ('hesaff_ptr = '
               'HESAFF_CLIB.new_hesaff_image(img_realpath, *hesaff_args)')
        print('msg = {!r}'.format(msg))
        print('hesaff_args = {!r}'.format(hesaff_args))
        raise
    return hesaff_ptr


#============================
# hesaff python interface
#============================


def get_hesaff_default_params():
    return HESAFF_PARAM_DICT.copy()


def get_is_debug_mode():
    return HESAFF_CLIB.is_debug_mode()


def get_cpp_version():
    r"""
    Returns:
        int: cpp_version

    CommandLine:
        python -m pyhesaff get_cpp_version

    Example:
        >>> # ENABLE_DOCTEST
        >>> from pyhesaff._pyhesaff import *  # NOQA
        >>> cpp_version = get_cpp_version()
        >>> isdebug = get_is_debug_mode()
        >>> print('cpp_version = %r' % (cpp_version,))
        >>> print('isdebug = %r' % (isdebug,))
        >>> assert cpp_version == 4, 'cpp version mimatch'
    """

    #str_ptr = HESAFF_CLIB.cmake_build_type()
    # copy c string into python
    #pystr = ctypes.c_char_p(str_ptr).value
    # need to free c string
    #HESAFF_CLIB.free_char(str_ptr)
    #print('pystr = %r' % (pystr,))
    #print('pystr = %s' % (pystr,))
    cpp_version = HESAFF_CLIB.get_cpp_version()
    return cpp_version


# full detection and extraction


def detect_feats(img_fpath, use_adaptive_scale=False, nogravity_hack=False, **kwargs):
    r"""
    driver function for detecting hessian affine keypoints from an image path.
    extra parameters can be passed to the hessian affine detector by using
    kwargs.

    Args:
        img_fpath (str): image file path on disk
        use_adaptive_scale (bool):
        nogravity_hack (bool):

    Kwargs:
        numberOfScales (int)         : default=3
        threshold (float)            : default=5.33333333333
        edgeEigenValueRatio (float)  : default=10.0
        border (int)                 : default=5
        maxIterations (int)          : default=16
        convergenceThreshold (float) : default=0.05
        smmWindowSize (int)          : default=19
        mrSize (float)               : default=5.19615242271
        spatialBins (int)            : default=4
        orientationBins (int)        : default=8
        maxBinValue (float)          : default=0.2
        initialSigma (float)         : default=1.6
        patchSize (int)              : default=41
        scale_min (float)            : default=-1.0
        scale_max (float)            : default=-1.0
        rotation_invariance (bool)   : default=False
        affine_invariance (bool)     : default=True

    Returns:
        tuple : (kpts, vecs)

    CommandLine:
        python -m pyhesaff detect_feats
        python -m pyhesaff detect_feats --show
        python -m pyhesaff detect_feats --show --fname star.png
        python -m pyhesaff detect_feats --show --fname zebra.png
        python -m pyhesaff detect_feats --show --fname astro.png
        python -m pyhesaff detect_feats --show --fname carl.jpg

        python -m pyhesaff detect_feats --show --fname astro.png --ri
        python -m pyhesaff detect_feats --show --fname astro.png --ai

        python -m pyhesaff detect_feats --show --fname easy1.png --no-ai
        python -m pyhesaff detect_feats --show --fname easy1.png --no-ai --numberOfScales=1 --verbose
        python -m pyhesaff detect_feats --show --fname easy1.png --no-ai --scale-max=100 --verbose
        python -m pyhesaff detect_feats --show --fname easy1.png --no-ai --scale-min=20 --verbose
        python -m pyhesaff detect_feats --show --fname easy1.png --no-ai --scale-min=100 --verbose
        python -m pyhesaff detect_feats --show --fname easy1.png --no-ai --scale-max=20 --verbose

        python -m vtool.test_constrained_matching visualize_matches --show
        python -m vtool.tests.dummy testdata_ratio_matches --show

        python -m pyhesaff detect_feats --show --fname easy1.png --ai \
            --verbose --scale-min=35 --scale-max=40

        python -m pyhesaff detect_feats --show --fname easy1.png --ai \
            --verbose --scale-min=35 --scale-max=40&
        python -m pyhesaff detect_feats --show --fname easy1.png --no-ai \
            --verbose --scale-min=35 --scale-max=40&
        python -m pyhesaff detect_feats --show --fname easy1.png --no-ai \
            --verbose --scale-max=40 --darken .5
        python -m pyhesaff detect_feats --show --fname easy1.png --no-ai \
            --verbose --scale-max=30 --darken .5
        python -m pyhesaff detect_feats --show --fname easy1.png --ai \
            --verbose --scale-max=30 --darken .5

        # DENSE KEYPOINTS
        python -m pyhesaff detect_feats --show --fname astro.png \
            --no-affine-invariance --numberOfScales=1 --maxPyramidLevels=1 \
            --use_dense --dense_stride=64
        python -m pyhesaff detect_feats --show --fname astro.png \
            --no-affine-invariance --numberOfScales=1 --maxPyramidLevels=1 \
            --use_dense --dense_stride=64 --rotation-invariance
        python -m pyhesaff detect_feats --show --fname astro.png \
            --affine-invariance --numberOfScales=1 --maxPyramidLevels=1 \
            --use_dense --dense_stride=64
        python -m pyhesaff detect_feats --show --fname astro.png \
            --no-affine-invariance --numberOfScales=3 \
            --maxPyramidLevels=2 --use_dense --dense_stride=32

        python -m pyhesaff detect_feats --show --only_count=False

    Example0:
        >>> # ENABLE_DOCTEST
        >>> # Test simple detect
        >>> from pyhesaff._pyhesaff import *  # NOQA
        >>> TAU = 2 * np.pi
        >>> img_fpath = grab_test_imgpath(ub.argval('--fname', default='astro.png'))
        >>> kwargs = argparse_hesaff_params()
        >>> print('kwargs = %r' % (kwargs,))
        >>> (kpts, vecs) = detect_feats(img_fpath, **kwargs)
        >>> # Show keypoints
        >>> # xdoctest: +REQUIRES(--show)
        >>> imgBGR = imread(img_fpath)
        >>> # take a random stample
        >>> frac = ub.argval('--frac', default=1.0)
        >>> print('frac = %r' % (frac,))
        >>> idxs = vecs[0:int(len(vecs) * frac)]
        >>> vecs, kpts = vecs[idxs], kpts[idxs]
        >>> default_showkw = dict(ori=False, ell=True, ell_linewidth=2,
        >>>                       ell_alpha=.4, ell_color='distinct')
        >>> print('default_showkw = %r' % (default_showkw,))
        >>> #showkw = ut.argparse_dict(default_showkw)
        >>> #import plottool as pt
        >>> #pt.interact_keypoints.ishow_keypoints(imgBGR, kpts, vecs, **showkw)
        >>> #pt.show_if_requested()
    """
    # Load image
    hesaff_ptr = _new_fpath_hesaff(img_fpath, **kwargs)
    # Get num detected
    nKpts = HESAFF_CLIB.detect(hesaff_ptr)
    # Allocate arrays
    kpts = alloc_kpts(nKpts)
    vecs = alloc_vecs(nKpts)
    # Populate arrays
    HESAFF_CLIB.exportArrays(hesaff_ptr, nKpts, kpts, vecs)
    HESAFF_CLIB.free_hesaff(hesaff_ptr)
    if use_adaptive_scale:  # Adapt scale if requested
        kpts, vecs = adapt_scale(img_fpath, kpts)
    if nogravity_hack:
        kpts, vecs = vtool_adapt_rotation(img_fpath, kpts)
    return kpts, vecs


def detect_feats2(img_or_fpath, **kwargs):
    """
    General way of detecting from either an fpath or ndarray

    Args:
        img_or_fpath (str or ndarray):  file path string

    Returns:
        tuple
    """
    if isinstance(img_or_fpath, str):
        fpath = img_or_fpath
        return detect_feats(fpath, **kwargs)
    else:
        img = img_or_fpath
        return detect_feats_in_image(img, **kwargs)


def detect_feats_list(image_paths_list, **kwargs):
    """
    Args:
        image_paths_list (list): A list of image paths

    Returns:
        tuple: (kpts_list, vecs_list) A tuple of lists of keypoints and decsriptors

    Kwargs:
        numberOfScales (int)         : default=3
        threshold (float)            : default=5.33333333333
        edgeEigenValueRatio (float)  : default=10.0
        border (int)                 : default=5
        maxIterations (int)          : default=16
        convergenceThreshold (float) : default=0.05
        smmWindowSize (int)          : default=19
        mrSize (float)               : default=5.19615242271
        spatialBins (int)            : default=4
        orientationBins (int)        : default=8
        maxBinValue (float)          : default=0.2
        initialSigma (float)         : default=1.6
        patchSize (int)              : default=41
        scale_min (float)            : default=-1.0
        scale_max (float)            : default=-1.0
        rotation_invariance (bool)   : default=False

    CommandLine:
        python -m pyhesaff._pyhesaff detect_feats_list --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from pyhesaff._pyhesaff import *  # NOQA
        >>> fpath = grab_test_imgpath('astro.png')
        >>> image_paths_list = [grab_test_imgpath('carl.jpg'), grab_test_imgpath('star.png'), fpath]
        >>> (kpts_list, vecs_list) = detect_feats_list(image_paths_list)
        >>> #print((kpts_list, vecs_list))
        >>> # Assert that the normal version agrees
        >>> serial_list = [detect_feats(fpath) for fpath in image_paths_list]
        >>> kpts_list2 = [c[0] for c in serial_list]
        >>> vecs_list2 = [c[1] for c in serial_list]
        >>> diff_kpts = [kpts - kpts2 for kpts, kpts2 in zip(kpts_list, kpts_list2)]
        >>> diff_vecs = [vecs - vecs2 for vecs, vecs2 in zip(vecs_list, vecs_list2)]
        >>> assert all([x.sum() == 0 for x in diff_kpts]), 'inconsistent results'
        >>> assert all([x.sum() == 0 for x in diff_vecs]), 'inconsistent results'
    """
    # Get Num Images
    num_imgs = len(image_paths_list)

    # Cast string list to C
    realpaths_list = [realpath(path).encode('ascii') for path in image_paths_list]

    c_strs = _cast_strlist_to_C(realpaths_list)

    # Get algorithm parameters
    hesaff_params = HESAFF_PARAM_DICT.copy()
    hesaff_params.update(kwargs)
    # pass all parameters to HESAFF_CLIB
    hesaff_args = hesaff_params.values()

    length_array = np.empty(num_imgs, dtype=int_t)
    detector_array = HESAFF_CLIB.detectFeaturesListStep1(num_imgs, c_strs, *hesaff_args)
    HESAFF_CLIB.detectFeaturesListStep2(num_imgs, detector_array, length_array)
    total_pts = length_array.sum()
    # allocate arrays
    total_num_arrays = num_imgs * total_pts
    flat_kpts_ptr = alloc_kpts(total_num_arrays)
    flat_vecs_ptr = alloc_vecs(total_num_arrays)
    # TODO: get this working
    offset_array = np.roll(length_array.cumsum(), 1).astype(int_t)
    # np.array([0] + length_array.cumsum().tolist()[0:-1], int_t)
    offset_array[0] = 0
    HESAFF_CLIB.detectFeaturesListStep3(num_imgs, detector_array,
                                         length_array, offset_array,
                                         flat_kpts_ptr, flat_vecs_ptr)

    # reshape into jagged arrays
    kpts_list = [flat_kpts_ptr[o:o + l] for o, l in zip(offset_array, length_array)]
    vecs_list = [flat_vecs_ptr[o:o + l] for o, l in zip(offset_array, length_array)]

    return kpts_list, vecs_list


def detect_feats_in_image(img, **kwargs):
    r"""
    Takes a preloaded image and detects keypoints and descriptors

    Args:
        img (ndarray[uint8_t, ndim=2]):  image data, should be in BGR or grayscale

    Returns:
        tuple: (kpts, vecs)

    CommandLine:
        python -m pyhesaff detect_feats_in_image --show
        python -m pyhesaff detect_feats_in_image --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from pyhesaff._pyhesaff import *  # NOQA
        >>> img_fpath = grab_test_imgpath('astro.png')
        >>> img = imread(img_fpath)
        >>> (kpts, vecs) = detect_feats_in_image(img)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool as pt
        >>> pt.interact_keypoints.ishow_keypoints(img, kpts, vecs, ori=True,
        >>>                                       ell_alpha=.4, color='distinct')
        >>> pt.set_figtitle('Detect Kpts in Image')
        >>> pt.show_if_requested()
    """
    #Valid keyword arguments are: + str(HESAFF_PARAM_DICT.keys())
    hesaff_ptr = _new_image_hesaff(img, **kwargs)
    # Get num detected
    nKpts = HESAFF_CLIB.detect(hesaff_ptr)
    # Allocate arrays
    kpts = alloc_kpts(nKpts)
    vecs = alloc_vecs(nKpts)
    HESAFF_CLIB.exportArrays(hesaff_ptr, nKpts, kpts, vecs)  # Populate arrays
    HESAFF_CLIB.free_hesaff(hesaff_ptr)
    return kpts, vecs


def detect_num_feats_in_image(img, **kwargs):
    r"""
    Just quickly returns how many keypoints are in the image. Does not attempt
    to return or store the values.

    It is a good idea to turn off things like ai and ri here.

    Args:
        img (ndarray[uint8_t, ndim=2]):  image data

    Returns:
        int: nKpts

    ISSUE: there seems to be an inconsistency for jpgs between this and detect_feats

    CommandLine:
        python -m pyhesaff detect_num_feats_in_image:0 --show
        python -m pyhesaff detect_num_feats_in_image:1 --show
        python -m xdoctest pyhesaff detect_num_feats_in_image:0

    Example:
        >>> # ENABLE_DOCTEST
        >>> from pyhesaff._pyhesaff import *  # NOQA
        >>> img_fpath = grab_test_imgpath('zebra.png')
        >>> img = imread(img_fpath)
        >>> nKpts = detect_num_feats_in_image(img)
        >>> kpts, vecs = detect_feats_in_image(img)
        >>> #assert nKpts == len(kpts), 'inconsistency'
        >>> result = ('nKpts = %s' % (ub.repr2(nKpts),))
        >>> print(result)

    Example1:
        >>> # TIMEDOCTEST
        >>> from pyhesaff._pyhesaff import *  # NOQA
        >>> setup = ub.codeblock(
            '''
            import pyhesaff
            img_fpath = grab_test_imgpath('carl.jpg')
            img = imread(img_fpath)
            ''')
        >>> stmt_list = [
        >>>    'pyhesaff.detect_feats_in_image(img)',
        >>>    'pyhesaff.detect_num_feats_in_image(img, affine_invariance=False)',
        >>>    'pyhesaff.detect_num_feats_in_image(img)',
        >>> ]
        >>> iterations = 30
        >>> verbose = True
        >>> #ut.timeit_compare(stmt_list, setup=setup, iterations=iterations,
        >>> #                  verbose=verbose, assertsame=False)

    """
    # We dont need to find vectors at all here
    kwargs['only_count'] = True
    #kwargs['only_count'] = False
    #Valid keyword arguments are: + str(HESAFF_PARAM_DICT.keys())
    hesaff_ptr = _new_image_hesaff(img, **kwargs)
    # Get num detected
    nKpts = HESAFF_CLIB.detect(hesaff_ptr)
    HESAFF_CLIB.free_hesaff(hesaff_ptr)
    return nKpts


# just extraction


def extract_vecs(img_fpath, kpts, **kwargs):
    r"""
    Extract SIFT descriptors at keypoint locations

    Args:
        img_fpath (str):
        kpts (ndarray[float32_t, ndim=2]):  keypoints

    Returns:
        ndarray[uint8_t, ndim=2]: vecs -  descriptor vectors

    CommandLine:
        python -m pyhesaff extract_vecs:0
        python -m pyhesaff extract_vecs:1 --fname=astro.png
        python -m pyhesaff extract_vecs:1 --fname=patsy.jpg --show
        python -m pyhesaff extract_vecs:1 --fname=carl.jpg
        python -m pyhesaff extract_vecs:1 --fname=zebra.png

    Example:
        >>> # ENABLE_DOCTEST
        >>> from pyhesaff._pyhesaff import *  # NOQA
        >>> img_fpath = grab_test_imgpath('carl.jpg')
        >>> kpts = np.array([[20, 25, 5.21657705, -5.11095951, 24.1498699, 0],
        >>>                  [29, 25, 2.35508823, -5.11095952, 24.1498692, 0],
        >>>                  [30, 30, 12.2165705, 12.01909553, 10.5286992, 0],
        >>>                  [31, 29, 13.3555705, 17.63429554, 14.1040992, 0],
        >>>                  [32, 31, 16.0527005, 3.407312351, 11.7353722, 0]], dtype=np.float32)
        >>> vecs = extract_vecs(img_fpath, kpts)
        >>> result = 'vecs = {}'.format(vecs)
        >>> print(result)

    Example:
        >>> # ENABLE_DOCTEST
        >>> from pyhesaff._pyhesaff import *  # NOQA
        >>> img_fpath = grab_test_imgpath(ub.argval('--fname', default='astro.png'))
        >>> # Extract original keypoints
        >>> kpts, vecs1 = detect_feats(img_fpath)
        >>> # Re-extract keypoints
        >>> vecs2 = extract_vecs(img_fpath, kpts)
        >>> # Descriptors should be the same
        >>> errors = (vecs1.astype(float) - vecs2.astype(float)).sum(axis=1)
        >>> errors_index = np.nonzero(errors)[0]
        >>> print('errors = %r' % (errors,))
        >>> print('errors_index = %r' % (errors_index,))
        >>> print('errors.sum() = %r' % (errors.sum(),))
        >>> # VISUALIZTION
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool as pt
        >>> # Extract the underlying grayscale patches
        >>> img = imread(img_fpath)
        >>> patch_list = extract_patches(img, kpts)
        >>> pt.interact_keypoints.ishow_keypoints(img_fpath, kpts[errors_index], vecs1[errors_index], fnum=1)
        >>> ax = pt.draw_patches_and_sifts(patch_list[errors_index], vecs1[errors_index], pnum=(1, 2, 1), fnum=2)
        >>> ax.set_title('patch extracted')
        >>> ax = pt.draw_patches_and_sifts(patch_list[errors_index], vecs2[errors_index], pnum=(1, 2, 2), fnum=2)
        >>> ax.set_title('image extracted')
        >>> pt.set_figtitle('Error Keypoints')
        >>> pt.show_if_requested()
    """
    hesaff_ptr = _new_fpath_hesaff(img_fpath, **kwargs)
    nKpts = len(kpts)
    # allocate memory for new decsriptors
    vecs = alloc_vecs(nKpts)
    # kpts might not be contiguous
    kpts = np.ascontiguousarray(kpts)
    # extract decsriptors at given locations
    HESAFF_CLIB.extractDesc(hesaff_ptr, nKpts, kpts, vecs)
    HESAFF_CLIB.free_hesaff(hesaff_ptr)
    return vecs


def extract_patches(img_or_fpath, kpts, **kwargs):
    r"""
    Extract patches used to compute SIFT descriptors.

    Args:
        img_or_fpath (ndarray or str):
        kpts (ndarray[float32_t, ndim=2]):  keypoints

    CommandLine:
        python -m pyhesaff extract_patches:0 --show
        python -m pyhesaff extract_vecs:1 --fname=astro.png
        python -m pyhesaff extract_vecs:1 --fname=patsy.jpg --show
        python -m pyhesaff extract_vecs:1 --fname=carl.jpg
        python -m pyhesaff extract_vecs:1 --fname=zebra.png

    Example:
        >>> # ENABLE_DOCTEST
        >>> # xdoctest: +REQUIRES(module:vtool)
        >>> from pyhesaff._pyhesaff import *  # NOQA
        >>> import vtool as vt
        >>> kwargs = {}
        >>> img_fpath = grab_test_imgpath('carl.jpg')
        >>> img = imread(img_fpath)
        >>> img_or_fpath = img
        >>> kpts, vecs1 = detect_feats(img_fpath)
        >>> kpts = kpts[1::len(kpts) // 9]
        >>> vecs1 = vecs1[1::len(vecs1) // 9]
        >>> cpp_patch_list = extract_patches(img, kpts)
        >>> py_patch_list_ = np.array(vt.get_warped_patches(img_or_fpath, kpts, patch_size=41)[0])
        >>> py_patch_list = np.array(vt.convert_image_list_colorspace(py_patch_list_, 'gray'))
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool as pt
        >>> ax = pt.draw_patches_and_sifts(cpp_patch_list, None, pnum=(1, 2, 1))
        >>> ax.set_title('C++ extracted')
        >>> ax = pt.draw_patches_and_sifts(py_patch_list, None, pnum=(1, 2, 2))
        >>> ax.set_title('Python extracted')
        >>> pt.show_if_requested()
    """
    if isinstance(img_or_fpath, str):
        hesaff_ptr = _new_fpath_hesaff(img_or_fpath, **kwargs)
    else:
        hesaff_ptr = _new_image_hesaff(img_or_fpath, **kwargs)
    nKpts = len(kpts)
    patch_list = alloc_patches(nKpts, 41)   # allocate memory for patches
    patch_list[:] = 0
    kpts = np.ascontiguousarray(kpts)  # kpts might not be contiguous
    # extract decsriptors at given locations
    HESAFF_CLIB.extractPatches(hesaff_ptr, nKpts, kpts, patch_list)
    HESAFF_CLIB.free_hesaff(hesaff_ptr)
    return patch_list


def extract_desc_from_patches(patch_list):
    r"""
    Careful about the way the patches are extracted here.

    Args:
        patch_list (ndarray[ndims=3]):

    CommandLine:
        python -m pyhesaff extract_desc_from_patches
        python -m pyhesaff extract_desc_from_patches  --show
        python -m pyhesaff extract_desc_from_patches:1 --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> # xdoctest: +REQUIRES(module:vtool)
        >>> from pyhesaff._pyhesaff import *  # NOQA
        >>> import vtool as vt
        >>> img_fpath = grab_test_imgpath(ub.argval('--fname', default='astro.png'))
        >>> # First extract keypoints normally
        >>> (orig_kpts_list, orig_vecs_list) = detect_feats(img_fpath)
        >>> # Take 9 keypoints
        >>> img = imread(img_fpath)
        >>> kpts_list = orig_kpts_list[1::len(orig_kpts_list) // 9]
        >>> vecs_list = orig_vecs_list[1::len(orig_vecs_list) // 9]
        >>> # Extract the underlying grayscale patches (using different patch_size)
        >>> patch_list_ = np.array(vt.get_warped_patches(img, kpts_list, patch_size=64)[0])
        >>> patch_list = np.array(vt.convert_image_list_colorspace(patch_list_, 'gray'))
        >>> # Extract descriptors from the patches
        >>> vecs_array = extract_desc_from_patches(patch_list)

    Example:
        >>> # ENABLE_DOCTEST
        >>> from pyhesaff._pyhesaff import *  # NOQA
        >>> img_fpath = grab_test_imgpath(ub.argval('--fname', default='astro.png'))
        >>> # First extract keypoints normally
        >>> (orig_kpts_list, orig_vecs_list) = detect_feats(img_fpath)
        >>> # Take 9 keypoints
        >>> img = imread(img_fpath)
        >>> kpts_list = orig_kpts_list[1::len(orig_kpts_list) // 9]
        >>> vecs_list = orig_vecs_list[1::len(orig_vecs_list) // 9]
        >>> # Extract the underlying grayscale patches
        >>> patch_list = extract_patches(img, kpts_list)
        >>> patch_list = np.round(patch_list).astype(np.uint8)
        >>> # Currently its impossible to get the correct answer
        >>> # TODO: allow patches to be passed in as float32
        >>> # Extract descriptors from those patches
        >>> vecs_array = extract_desc_from_patches(patch_list)
        >>> # Comparse to see if they are close to the original descriptors
        >>> errors = (vecs_list.astype(float) - vecs_array.astype(float)).sum(axis=1)
        >>> print('Errors: %r' % (errors,))
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool as pt
        >>> ax = pt.draw_patches_and_sifts(patch_list, vecs_array, pnum=(1, 2, 1))
        >>> ax.set_title('patch extracted')
        >>> ax = pt.draw_patches_and_sifts(patch_list, vecs_list, pnum=(1, 2, 2))
        >>> ax.set_title('image extracted')
        >>> pt.show_if_requested()
    """
    ndims = len(patch_list.shape)
    if ndims == 4 and patch_list.shape[-1] == 1:
        print('[pyhesaff] warning need to reshape patch_list')
        # need to remove grayscale dimension, maybe it should be included
        patch_list = patch_list.reshape(patch_list.shape[0:3])
    elif ndims == 4 and patch_list.shape[-1] == 3:
        assert False, 'cannot handle color images yet'
    assert patch_list.flags['C_CONTIGUOUS'], 'patch_list must be contiguous array'
    num_patches, patch_h, patch_w = patch_list.shape[0:3]
    assert patch_h == patch_w, 'must be square patches'
    vecs_array = alloc_vecs(num_patches)
    #vecs_array[:] = 0
    #print('vecs_array = %r' % (vecs_array,))
    # If the input array list is memmaped it is a good idea to process in chunks
    CHUNKS = isinstance(patch_list, np.memmap)
    if not CHUNKS:
        HESAFF_CLIB.extractDescFromPatches(num_patches, patch_h, patch_w,
                                           patch_list, vecs_array)
    else:
        chunksize = 2048
        _iter = range(num_patches // chunksize)
        _progiter = ub.ProgIter(_iter, desc='extracting sift chunk')
        for ix in _progiter:
            lx = ix * chunksize
            rx = (ix + 1) * chunksize
            patch_sublist = np.array(patch_list[lx:rx])
            sublist_size = rx - lx
            HESAFF_CLIB.extractDescFromPatches(sublist_size, patch_h, patch_w,
                                               patch_sublist,
                                               vecs_array[lx:rx])
        last_size = num_patches - rx
        if last_size > 0:
            lx = rx
            rx = lx + last_size
            patch_sublist = np.array(patch_list[lx:rx])
            sublist_size = rx - lx
            HESAFF_CLIB.extractDescFromPatches(sublist_size, patch_h, patch_w,
                                               patch_sublist,
                                               vecs_array[lx:rx])
    #print('vecs_array = %r' % (vecs_array,))
    return vecs_array

#============================
# other
#============================


def test_rot_invar():
    r"""
    CommandLine:
        python -m pyhesaff test_rot_invar --show
        python -m pyhesaff test_rot_invar --show --nocpp

        python -m vtool.tests.dummy testdata_ratio_matches --show --ratio_thresh=1.0 --rotation_invariance
        python -m vtool.tests.dummy testdata_ratio_matches --show --ratio_thresh=1.1 --rotation_invariance

    Example:
        >>> # DISABLE_DODCTEST
        >>> from pyhesaff._pyhesaff import *  # NOQA
        >>> test_rot_invar()
    """
    import cv2
    import vtool as vt
    import plottool as pt
    TAU = 2 * np.pi
    fnum = pt.next_fnum()
    NUM_PTS = 5  # 9
    theta_list = np.linspace(0, TAU, NUM_PTS, endpoint=False)
    nRows, nCols = pt.get_square_row_cols(len(theta_list), fix=True)
    next_pnum = pt.make_pnum_nextgen(nRows, nCols)
    # Expand the border a bit around star.png
    pad_ = 100
    img_fpath = grab_test_imgpath('star.png')
    img_fpath2 = vt.pad_image_ondisk(img_fpath, pad_, value=26)
    for theta in theta_list:
        print('-----------------')
        print('theta = %r' % (theta,))
        img_fpath = vt.rotate_image_ondisk(img_fpath2, theta, border_mode=cv2.BORDER_REPLICATE)
        if not ub.argflag('--nocpp'):
            (kpts_list_ri, vecs_list2) = detect_feats(img_fpath, rotation_invariance=True)
            kpts_ri = kpts_list_ri[0:2]
        (kpts_list_gv, vecs_list1) = detect_feats(img_fpath, rotation_invariance=False)
        kpts_gv = kpts_list_gv[0:2]
        # find_kpts_direction
        imgBGR = imread(img_fpath)
        kpts_ripy = vt.find_kpts_direction(imgBGR, kpts_gv, DEBUG_ROTINVAR=False)
        # Verify results stdout
        #print('nkpts = %r' % (len(kpts_gv)))
        #print(vt.kpts_repr(kpts_gv))
        #print(vt.kpts_repr(kpts_ri))
        #print(vt.kpts_repr(kpts_ripy))
        # Verify results plot
        pt.figure(fnum=fnum, pnum=next_pnum())
        pt.imshow(imgBGR)
        #if len(kpts_gv) > 0:
        #    pt.draw_kpts2(kpts_gv, ori=True, ell_color=pt.BLUE, ell_linewidth=10.5)
        ell = False
        rect = True
        if not ub.argflag('--nocpp'):
            if len(kpts_ri) > 0:
                pt.draw_kpts2(kpts_ri, rect=rect, ell=ell, ori=True,
                              ell_color=pt.RED, ell_linewidth=5.5)
        if len(kpts_ripy) > 0:
            pt.draw_kpts2(kpts_ripy, rect=rect, ell=ell,  ori=True,
                          ell_color=pt.GREEN, ell_linewidth=3.5)
    pt.set_figtitle('green=python, red=C++')
    pt.show_if_requested()


def vtool_adapt_rotation(img_fpath, kpts):
    """ rotation invariance in python """
    import vtool.patch as ptool
    import vtool.image as gtool
    imgBGR = gtool.imread(img_fpath)
    kpts2 = ptool.find_kpts_direction(imgBGR, kpts)
    vecs2 = extract_vecs(img_fpath, kpts2)
    return kpts2, vecs2


def adapt_scale(img_fpath, kpts):
    import vtool.ellipse as etool
    nScales = 16
    nSamples = 16
    low, high = -1, 2
    kpts2 = etool.adaptive_scale(img_fpath, kpts, nScales, low, high, nSamples)
    # passing in 0 orientation results in gravity vector direction keypoint
    vecs2 = extract_vecs(img_fpath, kpts2)
    return kpts2, vecs2


# del type_, key, val

if __name__ == '__main__':
    """
    CommandLine:
        python -m pyhesaff._pyhesaff
        python -m pyhesaff._pyhesaff --allexamples
        python -m pyhesaff._pyhesaff --allexamples --noface --nosrc
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



#!/usr/bin/env python
from __future__ import absolute_import, division, print_function
import numpy as np
import utool as ut


def double_detect(img_fpath, **kw):
    import pyhesaff
    # Checks to make sure computation is determinsitc
    _kpts, _vecs = pyhesaff.detect_feats(img_fpath, **kw)
    kpts_, vecs_ = pyhesaff.detect_feats(img_fpath, **kw)
    assert np.all(kpts_ == _kpts)
    assert np.all(vecs_ == _vecs)
    print('double detect ok')
    return kpts_, vecs_


def test_ori_extract_main():
    """
    CommandLine:
        python -m pyhesaff.tests.test_exhaustive_ori_extract --test-test_ori_extract_main
        python -m pyhesaff.tests.test_exhaustive_ori_extract --test-test_ori_extract_main --show

    # Example:
    #     >>> # GUI_DOCTEST
    #     >>> from pyhesaff.tests.test_exhaustive_ori_extract import *  # NOQA
    #     >>> test_ori_extract_main()
    #     >>> ut.show_if_requested()
    """
    import pytest
    pytest.skip('Broken CI')

    import pyhesaff
    from plottool import draw_func2 as df2
    from plottool.viz_keypoints import show_keypoints
    import vtool  # NOQA
    import vtool.image as gtool
    import vtool.keypoint as ktool
    np.set_printoptions(threshold=5000, linewidth=5000, precision=3)
    # Read data
    print('[rotinvar] loading test data')

    img_fpath = ut.grab_test_imgpath('jeff.png')
    imgL = gtool.cvt_BGR2L(gtool.imread(img_fpath))
    detect_kw0 = {
    }
    detect_kw1 = {
        'scale_min': 20,
        'scale_max': 100
    }
    detect_kw2 = {
        'scale_min': 40,
        'scale_max': 60
    }
    detect_kw3 = {
        'scale_min': 45,
        'scale_max': 49
    }
    # Remove skew and anisotropic scaling
    def force_isotropic(kpts):
        kpts_ = kpts.copy()
        kpts_[:, ktool.SKEW_DIM] = 0
        kpts_[:, ktool.SCAX_DIM] = kpts_[:, ktool.SCAY_DIM]
        vecs_ = pyhesaff.extract_vecs(img_fpath, kpts_)
        return kpts_, vecs_

    def force_ori(kpts, ori):
        kpts_ = kpts.copy()
        kpts_[:, ktool.ORI_DIM] = ori
        vecs_ = pyhesaff.extract_vecs(img_fpath, kpts_)
        return kpts_, vecs_

    def shift_kpts(kpts, x, y):
        kpts_ = kpts.copy()
        kpts_[:, ktool.XDIM] += x
        kpts_[:, ktool.YDIM] += y
        vecs_ = pyhesaff.extract_vecs(img_fpath, kpts_)
        return kpts_, vecs_

    # --- Experiment ---
    kpts0, vecs0 = double_detect(img_fpath, **detect_kw0)
    kpts1, vecs1 = double_detect(img_fpath, **detect_kw1)
    kpts2, vecs2 = double_detect(img_fpath, **detect_kw2)
    #
    kpts3, vecs3 = double_detect(img_fpath, **detect_kw3)
    kpts4, vecs4 = force_isotropic(kpts3)
    kpts5, vecs5 = force_ori(kpts3, 1.45)
    kpts6, vecs6 = shift_kpts(kpts5, -60, -50)
    kpts7, vecs7 = force_ori(kpts6, 0)
    kpts8, vecs8 = force_ori(kpts7, 2.40)
    kpts9, vecs9 = force_ori(kpts8, 5.40)
    kpts10, vecs10 = force_ori(kpts9, 10.40)

    # --- Print ---

    # ---- Draw ----
    nRow, nCol = 1, 2
    df2.figure(fnum=2, doclf=True, docla=True)
    df2.figure(fnum=1, doclf=True, docla=True)

    def show_kpts_(fnum, pnum, kpts, vecs, title):
        print('--------')
        print('show_kpts: %r.%r' % (fnum, pnum))
        print('kpts  = %r' % (kpts,))
        print('scales = %r' % ktool.get_scales(kpts))
        # FIXME: this exists in ibeis. move to vtool
        #dev_consistency.check_vecs(vecs3)

        show_keypoints(imgL, kpts, sifts=vecs, pnum=pnum, rect=True,
                       ori=True, fnum=fnum, title=title, ell_alpha=1)

    show_kpts_(1, (nRow, nCol, 1), kpts3, vecs3, 'kpts3: original')
    show_kpts_(1, (nRow, nCol, 2), kpts4, vecs4, 'kpts4: isotropic + redetect')
    show_kpts_(2, (2, 3, 1), kpts5, vecs5, 'kpts5: force_ori + redetect')
    show_kpts_(2, (2, 3, 2), kpts6, vecs6, 'kpts6: shift')
    show_kpts_(2, (2, 3, 3), kpts7, vecs7, 'kpts7: shift + reorient')
    show_kpts_(2, (2, 3, 4), kpts8, vecs8, 'kpts8: shift + reorient')
    show_kpts_(2, (2, 3, 5), kpts9, vecs9, 'kpts9: reorient')
    show_kpts_(2, (2, 3, 6), kpts10, vecs10, 'kpts10: reorient')
    #df2.iup()

    #pinteract.interact_keypoints(imgBGR, kpts2, vecs, arrow=True, rect=True)
    #exec(df2.present(wh=800))


if __name__ == '__main__':
    """
    CommandLine:
        python -m pyhesaff.tests.test_exhaustive_ori_extract
        python -m pyhesaff.tests.test_exhaustive_ori_extract --allexamples
        python -m pyhesaff.tests.test_exhaustive_ori_extract --allexamples --noface --nosrc
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



# See vtool dev/build_base_docker2.py
# See ~/code/vtool_ibeis/dev/build_base_docker2.py

"""
python ~/code/vtool_ibeis/dev/build_base_docker2.py --arch=x86_64 --parent_image_prefix=manylinux2014 --opencv
python ~/code/vtool_ibeis/dev/build_base_docker2.py --arch=i686 --parent_image_prefix=manylinux2014 --opencv
"""



#!/usr/bin/env python
from __future__ import absolute_import, division, print_function
import numpy as np
import itertools as it
from six.moves import range
import six


px_ = 0


def test_draw_keypoint_main():
    r"""
    CommandLine:
        python -m pyhesaff.tests.test_draw_keypoint --test-test_draw_keypoint_main --show

    # Example:
    #     >>> # DISABLE_DOCTEST
    #     >>> from pyhesaff.tests.test_draw_keypoint import *  # NOQA
    #     >>> test_draw_keypoint_main()
    #     >>> ut.show_if_requested()
    """
    import pytest
    pytest.skip('Broken in CI')

    from plottool import draw_func2 as df2
    from plottool import mpl_keypoint
    import vtool.keypoint as ktool  # NOQA

    # TODO: Gui tests yield:
    # Jul 13 13:14:53 www.longerdog.com Python[23974] <Error>: This user is not allowed access to the window system right now.
    # don't do window access without --show

    TAU = 2 * np.pi

    # Hack these directions to be relative to gravity
    #RIGHT = ((0 * TAU / 4) - ktool.GRAVITY_THETA) % TAU
    DOWN  = ((1 * TAU / 4) - ktool.GRAVITY_THETA) % TAU
    #LEFT  = ((2 * TAU / 4) - ktool.GRAVITY_THETA) % TAU
    #UP    = ((3 * TAU / 4) - ktool.GRAVITY_THETA) % TAU

    def test_keypoint(xscale=1, yscale=1, ori=DOWN, skew=0):
        # Test Keypoint
        x, y = 0, 0
        iv11, iv21, iv22 = xscale, skew, yscale
        kp = np.array([x, y, iv11, iv21, iv22, ori])

        # Test SIFT descriptor
        sift = np.zeros(128)
        sift[ 0: 8]   = 1
        sift[ 8:16]   = .5
        sift[16:24]   = .0
        sift[24:32]   = .5
        sift[32:40]   = .8
        sift[40:48]   = .8
        sift[48:56]   = .1
        sift[56:64]   = .2
        sift[64:72]   = .3
        sift[72:80]   = .4
        sift[80:88]   = .5
        sift[88:96]   = .6
        sift[96:104]  = .7
        sift[104:112] = .8
        sift[112:120] = .9
        sift[120:128] = 1
        sift = sift / np.sqrt((sift ** 2).sum())
        sift = np.round(sift * 255)

        kpts = np.array([kp])
        sifts = np.array([sift])
        return kpts, sifts

    def square_axis(ax, s=3):
        ax.set_xlim(-s, s)
        ax.set_ylim(-s, s)
        ax.set_aspect('equal')
        ax.invert_yaxis()
        df2.set_xticks([])
        df2.set_yticks([])

    def test_shape(ori=0, skew=0, xscale=1, yscale=1, pnum=(1, 1, 1), fnum=1):
        df2.figure(fnum=fnum, pnum=pnum)
        kpts, sifts = test_keypoint(ori=ori, skew=skew, xscale=xscale, yscale=yscale)
        ax = df2.gca()
        square_axis(ax)
        mpl_keypoint.draw_keypoints(ax, kpts, sifts=sifts, ell_color=df2.ORANGE, ori=True,
                                    rect_color=df2.DARK_RED,
                                    ori_color=df2.DEEP_PINK, eig_color=df2.PINK,
                                    rect=True, eig=True, bin_color=df2.RED,
                                    arm1_color=df2.YELLOW, arm2_color=df2.BLACK)

        kptsstr = '\n'.join(ktool.get_kpts_strs(kpts))
        #print(kptsstr)
        df2.upperleft_text(kptsstr)

        title = 'xyscale=(%.1f, %.1f),\n skew=%.1f, ori=%.2ftau' % (xscale, yscale, skew, ori / TAU)
        df2.set_title(title)
        df2.dark_background()
        return kpts, sifts

    np.set_printoptions(precision=3)

    #THETA1 = DOWN
    #THETA2 = (DOWN + DOWN + RIGHT) / 3
    #THETA3 = (DOWN + RIGHT) / 2
    #THETA4 = (DOWN + RIGHT + RIGHT) / 3
    #THETA5 = RIGHT

    nRows = 2
    nCols = 4

    import plottool as pt
    #pnum_ = pt.pnum_generator(nRows, nCols).next
    pnum_ = pt.pnum_generator(nRows, nCols)

    #def pnum_(px=None):
    #    global px_
    #    if px is None:
    #        px_ += 1
    #        px = px_
    #    return (nRows, nCols, px)
    import ubelt as ub
    # ub.argval

    # MIN_ORI = ut.get_argval('--min-ori', float, DOWN)
    # MAX_ORI = ut.get_argval('--max-ori', float, DOWN + TAU - .2)
    # MIN_SWEW = ut.get_argval('--min-skew', float, 0)
    # MAX_SKEW = ut.get_argval('--max-skew', float, 1)
    MIN_ORI = float(ub.argval('--min-ori', DOWN))
    MAX_ORI = float(ub.argval('--max-ori', DOWN + TAU - .2))
    MIN_X = .5
    MAX_X = 2

    MIN_SWEW = float(ub.argval('--min-skew', 0))
    MAX_SKEW = float(ub.argval('--max-skew', 1))

    MIN_Y = .5
    MAX_Y = 2

    #kp_list = []

    for row, col in it.product(range(nRows), range(nCols)):
        #print((row, col))
        alpha = col / (nCols)
        beta  = row / (nRows)
        xsca = (MIN_X    * (1 - alpha)) + (MAX_X    * (alpha))
        ori  = (MIN_ORI  * (1 - alpha)) + (MAX_ORI  * (alpha))
        skew = (MIN_SWEW * (1 - beta))  + (MAX_SKEW * (beta))
        ysca = (MIN_Y    * (1 - beta))  + (MAX_Y    * (beta))

        kpts, sifts = test_shape(pnum=six.next(pnum_),
                                 ori=ori,
                                 skew=skew,
                                 xscale=xsca,
                                 yscale=ysca)
        #print('+----')
        #kp_list.append(kpts[0])
        #S_list = ktool.ge t_xy_axis_extents(kpts)
        #print('xscale=%r yscale=%r, skew=%r' % (xsca, ysca, skew))
        #print(S_list)

    #scale_factor = 1
    #offset = (0, 0)
    #(_xs, _ys, _iv11s, _iv12s, _iv21s, _iv22s, _oris) = ktool.scaled_kpts(kpts, scale_factor, offset)

    #invVR_aff2Ds = mpl_keypoint.get_invV_aff2Ds(_xs, _ys, _iv11s, _iv12s, _iv21s, _iv22s)
    #aff = invVR_aff2Ds[0]

    #ori = _oris[0]
    #aff2 = mpl.transforms.Affine2D().rotate(-ori)

    #print((aff + aff2).frozen())
    #print((aff2 + aff).frozen())

    #mpl_sift.draw_sift(ax, sift)
    #df2.update()

    #if ut.show_was_requested():
    #    exec(df2.present(wh=(900, 950)), globals(), locals())


if __name__ == '__main__':
    """
    CommandLine:
        python ~/code/pyhesaff/tests/test_draw_keypoint.py
    """
    import xdoctest
    xdoctest.doctest_module(__file__)
